#!env python
# -*- coding: iso-8859-1 -*-
###########################################################################
# 
# File:            sourceModbus.py
#
# License:         Donationware, see attached LICENSE file for more 
#                  information
#
# Author:          Olli Lammi (olammi@iki.fi)
#
# Version:         1.6e
#
# Date:            14.01.2015
#
# Description:     TaloLogger configuration wrapper for Modbus source.
#                  
# Version history: ** 14.01.2015 v1.6e (Olli Lammi) **
#                  Added to taloLogger. 
#
###########################################################################

# Imports

import string

from modules.core import log
from modules.core import configuration
from modules.core import dataSource

from modules.utils import modbus
from modules.utils import binary


###########################################################################

# Classes

class ModbusConf(log.Logging, configuration.Configurable, dataSource.DataSource):
    def __init__(self, modname):
        log.Logging.__init__(self, 'ModbusConf')
        configuration.Configurable.__init__(self)
        self.setModuleName(modname)
        dataSource.DataSource.__init__(self)
        self.setID(modname)

        self.modbusClient = None
        self.modbusType = None
        self.dataPoints = {}
        self.portName = None
        self.hostName = None
        self.tcpPort = 502
        self.baudRate = 19200
        self.parity = 'EVEN'
        self.stopBits = '1'
        self.dataBits = '8'
        self.resTimeout = 2.0
        
    def getServiceName(self):
        return self.getModuleName()

    def handleConfiguration(self, conf):
        self.modbusType = string.upper(conf.getValue('TYPE', 'RTU', self.getModuleName()))
        if not self.modbusType in ['RTU', 'ASCII', 'TCP']:
            return (-1, 'Invalid TYPE configuration')
        
        self.portName = conf.getValue('PORTNAME', '', self.getModuleName())
        self.hostName = conf.getValue('HOSTNAME', '', self.getModuleName())
        try: 
            self.tcpPort = int(conf.getValue('TCPPORT', '502', self.getModuleName()))
        except:
            return (-1, 'Invalid TCPPORT configuration')
        try: 
            self.baudRate = int(conf.getValue('BAUDRATE', '19200', self.getModuleName()))
        except:
            return (-1, 'Invalid BAUDRATE configuration')
        
        self.parity = string.upper(conf.getValue('PARITY', 'EVEN', self.getModuleName()))
        if not self.parity in ['E', 'O', 'N', 'M', 'S', 'EVEN', 'ODD', 'NONE', 'MARK', 'SPACE']:
            return (-1, 'Invalid PARITY configuration')
         
        self.stopBits = conf.getValue('STOPBITS', '1', self.getModuleName())
        if not self.stopBits in ['1', '2', '1.5']:
            return (-1, 'Invalid STOPBITS configuration')
        
        self.dataBits = conf.getValue('DATABITS', '8', self.getModuleName())
        if not self.dataBits in ['5', '6', '7', '8']:
            return (-1, 'Invalid DATABITS configuration')

        try: 
            self.resTimeout = float(conf.getValue('RESPONSETIMEOUT', '2.0', self.getModuleName()))
        except:
            return (-1, 'Invalid RESPONSETIMEOUT configuration')
        
        pointmap = {}
        
        for item in conf.getValue('COILSTATE', [], self.getModuleName()):
            temps = string.split(item, ':', 3)
            if len(temps) < 3:
                self.Log("ERROR: Invalid COILSTATE configuration: " + item)
                return (-1, 'Invalid COILSTATE configuration')
            else:
                try:
                    key = string.strip(temps[0])
                    unitid = int(string.strip(temps[1]))
                    address = int(string.strip(temps[2]))
                except:
                    key = ''
                if len(key) <= 0:
                    self.Log("ERROR: Invalid COILSTATE configuration: " + item)
                    return (-1, 'Invalid COILSTATE configuration')
                else:
                    tempitem = {}
                    tempitem['type'] = 'COIL'
                    tempitem['unitid'] = unitid
                    tempitem['address'] = address
                    tempitem['valuetype'] = 'bool'
                    tempitem['factor'] = 1
                    pointmap[key] = tempitem

        for item in conf.getValue('INPUTSTATE', [], self.getModuleName()):
            temps = string.split(item, ':', 3)
            if len(temps) < 3:
                self.Log("ERROR: Invalid INPUTSTATE configuration: " + item)
                return (-1, 'Invalid INPUTSTATE configuration')
            else:
                try:
                    key = string.strip(temps[0])
                    unitid = int(string.strip(temps[1]))
                    address = int(string.strip(temps[2]))
                except:
                    key = ''
                if len(key) <= 0:
                    self.Log("ERROR: Invalid INPUTSTATE configuration: " + item)
                    return (-1, 'Invalid INPUTSTATE configuration')
                else:
                    tempitem = {}
                    tempitem['type'] = 'IN'
                    tempitem['unitid'] = unitid
                    tempitem['address'] = address
                    tempitem['valuetype'] = 'bool'
                    tempitem['factor'] = 1
                    pointmap[key] = tempitem

        for item in conf.getValue('HOLDINGREGISTER', [], self.getModuleName()):
            temps = string.split(item, ':', 5)
            if len(temps) < 5:
                self.Log("ERROR: Invalid HOLDINGREGISTER configuration: " + item)
                return (-1, 'Invalid HOLDINGREGISTER configuration')
            else:
                try:
                    key = string.strip(temps[0])
                    unitid = int(string.strip(temps[1]))
                    address = int(string.strip(temps[2]))
                    valuetype = string.lower(string.strip(temps[3]))
                    factor = float(string.strip(temps[4]))
                except:
                    key = ''
                if len(key) <= 0:
                    self.Log("ERROR: Invalid HOLDINGREGISTER configuration: " + item)
                    return (-1, 'Invalid HOLDINGREGISTER configuration')
                else:
                    if not valuetype in ['int','uint','float','ufloat','bool']:
                        self.Log("ERROR: Invalid HOLDINGREGISTER configuration: " + item)
                        return (-1, 'Invalid HOLDINGREGISTER configuration')
                                        
                    tempitem = {}
                    tempitem['type'] = 'HOLDREG'
                    tempitem['unitid'] = unitid
                    tempitem['address'] = address
                    tempitem['valuetype'] = valuetype
                    tempitem['factor'] = factor
                    pointmap[key] = tempitem

        for item in conf.getValue('INPUTREGISTER', [], self.getModuleName()):
            temps = string.split(item, ':', 5)
            if len(temps) < 5:
                self.Log("ERROR: Invalid INPUTREGISTER configuration: " + item)
                return (-1, 'Invalid INPUTREGISTER configuration')
            else:
                try:
                    key = string.strip(temps[0])
                    unitid = int(string.strip(temps[1]))
                    address = int(string.strip(temps[2]))
                    valuetype = string.lower(string.strip(temps[3]))
                    factor = float(string.strip(temps[4]))
                except:
                    key = ''
                if len(key) <= 0:
                    self.Log("ERROR: Invalid INPUTREGISTER configuration: " + item)
                    return (-1, 'Invalid INPUTREGISTER configuration')
                else:
                    if not valuetype in ['int','uint','float','ufloat','bool']:
                        self.Log("ERROR: Invalid INPUTREGISTER configuration: " + item)
                        return (-1, 'Invalid INPUTREGISTER configuration')
                                        
                    tempitem = {}
                    tempitem['type'] = 'INREG'
                    tempitem['unitid'] = unitid
                    tempitem['address'] = address
                    tempitem['valuetype'] = valuetype
                    tempitem['factor'] = factor
                    pointmap[key] = tempitem

        self.dataPoints = pointmap
            
        return (1, '')
            
    def initConfiguration(self):        
        if self.modbusType == 'TCP':
            self.modbusClient = modbus.ModbusClientTCP(self.hostName, self.tcpPort)
        elif self.modbusType == 'RTU':
            self.modbusClient = modbus.ModbusClientSerial(self.portName, modbus.MODBUS_SERIAL_RTU)
        elif self.modbusType == 'ASCII':
            self.modbusClient = modbus.ModbusClientSerial(self.portName, modbus.MODBUS_SERIAL_ASCII)
        else:
            return (0, 'Invalid Modbus protocol type.' + self.modbusType)
        
        # set serial parameters
        if self.modbusType in ['RTU', 'ASCII']:
            self.modbusClient.setBaudrate(self.baudRate)
            self.modbusClient.setParity(self.parity)
            self.modbusClient.setDataBits(self.dataBits)
            self.modbusClient.setStopBits(self.stopBits)
        
        self.modbusClient.setResponseTimeout(self.resTimeout)
                
        return (1, '')

    def releaseConfiguration(self):
        if self.modbusClient != None:
            self.modbusClient.close()
            self.modbusClient = None
        return (1, '')

    def runDataSourceQueryCommandImpl(self, cmds):
        dataresult = {}
        if not self.modbusClient.open():
            self.Log("ERROR: Cannot open modbus connection.")

        for cmd in cmds:
            if not self.dataPoints.has_key(cmd):
                self.Log("ERROR: Unconfigured datapoint: %s" % cmd)
                dataresult[cmd] = ''
                continue
            
            dp = self.dataPoints[cmd]
            
            tempres = {}
            if dp['type'] == 'COIL':
                tempres = self.modbusClient.readCoils(dp['unitid'], dp['address'], 1)
            elif dp['type'] == 'IN':
                tempres = self.modbusClient.readDiscreteInputs(dp['unitid'], dp['address'], 1)
            elif dp['type'] == 'HOLDREG':
                tempres = self.modbusClient.readHoldingRegisters(dp['unitid'], dp['address'], 1)
            elif dp['type'] == 'INREG':
                tempres = self.modbusClient.readInputRegisters(dp['unitid'], dp['address'], 1)

            if tempres.has_key(dp['address']):
                self.Debug("Received data for unitid %d address %d: %s" % (dp['unitid'], dp['address'], tempres[dp['address']]))
                dataresult[cmd] = handleResultValue(tempres[dp['address']], dp['valuetype'], dp['factor'])
            else:
                self.Log("ERROR: Modbus query result does not contain value for unitid %d address %d" % (dp['unitid'], dp['address']))
                dataresult[cmd] = ''
        
        self.modbusClient.close()
        return dataresult

    @staticmethod
    def getModuleTypeName():
        return 'MODBUS'

    @staticmethod
    def getAllowedConfigurationKeys():
        return (['TYPE', 'PORTNAME', 'HOSTNAME', 'TCPPORT', 'BAUDRATE', 'PARITY', 'STOPBITS', 'DATABITS', 'RESPONSETIMEOUT'], ['COILSTATE', 'HOLDINGREGISTER', 'INPUTREGISTER', 'INPUTSTATE'])

# Functions

def handleResultValue(value, type, factor):
    if type == 'uint':
        return "%d" % int(value * factor)
    elif type == 'int':
        return "%d" % int(binary.twosComplementToInt(value, 16) * factor)
    elif type == 'ufloat':
        return "%f" % (float(value) * factor)
    elif type == 'float':
        return "%f" % (float(binary.twosComplementToInt(value, 16)) * factor)
    else:
        return "%d" % value
