#!env python
# -*- coding: iso-8859-1 -*-
###########################################################################
# 
# File:            nibeRS485Serial.py
#
# License:         Donationware, see attached LICENSE file for more 
#                  information
#
# Author:          Olli Lammi (olammi@iki.fi)
#                  (Nibe datapoint table from OpenHab Nibe binding
#                   subproject by Pauli Anttila) 
#
# Version:         0.1a
#
# Date:            16.01.2015
#
# Description:     Library to communicate with the Nibe heat pump controllers
#                  using RS485 adapter connected to Nibe bus or OpenHab
#                  NibeGW producing UDP frame packets.
#                  
# Supported devices: 
#                  TBD
#
# Requirements:    Python interpreter 2.6 or newer (www.python.org)
# 
#                  Python Serial Port Extension: pySerial  
#                  (http://pyserial.wiki.sourceforge.net/pySerial)
#
# Version history: ** 16.01.2015 v0.1a (Olli Lammi) **
#                  First test version. 
#
###########################################################################

# Imports

import sys, os 
import string
import time 
import socket

from modules.core import threads
from modules.core import log
from modules.utils import lockfile
from modules.utils import binary

# pySerial module, not loaded until init of NibeRS485Serial class instance
serial = None

###########################################################################

# Constants

BAUDRATE = 9600
TIMEOUT = 2
WRTIMEOUT = 2
DATAVALID = 240
SLEEP_AFTER_FAIL = 10
UDP_RECEIVE_BUFFER_SIZE = 2048

###########################################################################

# Device configuration data

TYPE_INT8 = 1 
TYPE_INT16 = 2
TYPE_INT32 = 3 
TYPE_UINT8 = 4 
TYPE_UINT16 = 5
TYPE_UINT32 = 6
TYPE_INT16_10 = 7
TYPE_INT8_10 = 8

NIBE_DEVICES = { 'DEFAULT': [ [40004, 'BT1 outdoor temp', TYPE_INT16_10] ] }

 
###########################################################################

# Classes

class NibeRS485Base(threads.Thread, log.Logging):
    def __init__(self, device):
        threads.Thread.__init__(self)
        log.Logging.__init__(self, 'NibeRS485')

        self.NIBE_DEVICE = device
                
        self.data = {}
        self.data_lock = threads.Lock()

    def setDevice(self, device):
        self.NIBE_DEVICE = device
                        
    def handleBuffer(self, buff):
        # check that the 4th byte is correct command 0x68
        if len(buff) < 4 or buff[3] != '\x68':
            self.Debug("Ignoring frame with unknown command in byte 4.")
            return
        
        temp = getNibeDataPart(buff)
        temp = fixNibeDataPart(temp)
        if self.data_lock.lock_wait():
            id = 0
            l = len(temp)
            i = 0
            while i < l:
                id = (ord(temp[i+1]) << 8) + ord(temp[i])
                i = i + 2
                value = (ord(temp[i+1]) << 8) + ord(temp[i])
                i = i + 2
                self.data[id] = [time.time(), value]
                self.Debug("Got data for id %d: 0x%04X" % (id, value))
            self.data_lock.free()
        else:
            self.Log("Cannot get data lock.")
            
    def runQueryCommand(self, cmd):
        if self.hasTerminated() and self.isFailed():
            if not self.startModule():
                self.Log("Could not restart module after being failed.")
            else:
                self.Log("Module restarted after being failed.")
        
        id = 0
        type = 0
        for t in NIBE_DEVICES[self.NIBE_DEVICE]:
            if t[1] == cmd:
                id = t[0]
                type = t[2]
                break

        if type == 0:
            self.Log("ERROR: Invalid NIBE key: " + cmd)
            return ""

        res = self.runQueryId(id)
        if res != None:
            return convertNibeMessage(type, res)
        return ""

    def runQueryId(self, id):
        self.Debug("Running nibeBus query with id " + repr(id))

        res = None
        if self.data_lock.lock_wait():
            if self.data.has_key(id):
                temp = self.data[id]
                if (time.time() - temp[0]) <= DATAVALID:
                    res = temp[1]
                else:
                    self.Debug("Data expired with id " + repr(id))
            else:
                self.Debug("No data found with id " + repr(id))
            self.data_lock.free()
        else:
            self.Log("Cannot get data lock.")
        return res   

    def startModule(self):
        return False


class NibeRS485Serial(NibeRS485Base):
    def __init__(self, port, device):
        NibeRS485Base.__init__(self, device)
       
        global serial
        serial = __import__('serial') 
          
        self.SERPORT = port
        self.serio = 0
                    
        self.LOCK = None
        
    def setPort(self, port):
        self.SERPORT = port
                
    def run(self):        
        try:              
            while self.isRunning():
                if not self.isOpen():
                    if not self.openPort():
                        self.Log("Cannot open serial port.")
                        self.setFail()
                        break
                    
                self.Debug("Flushing serial input buffer.")
                temp = ' '
                while self.isRunning() and len(temp) > 0 and self.serio.inWaiting() > 0:
                    temp = self.serio.read(1)

                res = ''
                prevchar = ''
                while self.isRunning():
                    temp = self.serio.read(1)
                    if len(temp) > 0:
                        res = res + temp
                                        
                        stat = 1
                        while self.isRunning() and stat != 0 and len(res) > 0:                    
                            (stat, flen) = checkNibeMessage(res, prevchar)
                            if stat > 0:
                                rcvdata = res[:flen]
                                prevchar = rcvdata[flen-1]
                                res = res[flen:]
                                if stat == 1:
                                    self.serio.write('\x06')
                                    self.serio.flush()
                                    self.Debug("Received:\n" + log.dumpBuffer(rcvdata))
                                    self.Debug("Sending ACK")
                                    self.handleBuffer(rcvdata)
                                elif stat == 2:
                                    self.serio.write('\x15')
                                    self.serio.flush()
                                    self.Debug("Received:\n" + log.dumpBuffer(rcvdata))
                                    self.Debug("Sending NAK")
                                else:
                                    self.Debug("Received:\n" + log.dumpBuffer(rcvdata))
                                    self.Debug("Ignoring")                                    
                                                                
                            elif stat == -1:
                                prevchar = res[0]
                                res = res[1:]

                self.closePort()
                                
                if self.isRunning():
                    time.sleep(SLEEP_AFTER_FAIL)

        except Exception, e:
            self.Log("Exception: " + e.__str__())
            self.setFail()
        except IOError, ioe:
            self.Log("IOError: " + ioe.__str__())
            self.setFail()

        self.closePort()
       
        self.Log("Nibe Bus serial thread stopped.")
    
                
    def openPort(self):
        if not NIBE_DEVICES.has_key(self.NIBE_DEVICE):
            self.Log("ERROR: Invalid NIBE device type: " + self.NIBE_DEVICE)
            return 0

        if not self.LOCK.lock():
            self.Log("ERROR: Unable to aquire lockfile for nibeBus serial port: " + self.LOCK.getName())
            return 0

        try:
            self.serio = serial.Serial(port=self.SERPORT,baudrate=BAUDRATE, \
                                bytesize=serial.EIGHTBITS, \
                                parity=serial.PARITY_NONE, stopbits=serial.STOPBITS_ONE, \
                                xonxoff=0, rtscts=0, dsrdtr=0, timeout=TIMEOUT, writeTimeout=WRTIMEOUT)
            if not self.serio.isOpen():
                self.serio = 0
                self.LOCK.free()
                return 0 
        except:
            self.Log("Error opening nibeBus serial port: " + self.SERPORT)
            self.serio = 0
            self.LOCK.free()
            return 0
        
        self.Debug("Opened nibeBus serial port: " + self.SERPORT)
        return 1
        
    def closePort(self):
        if self.serio:
            try:
                self.serio.close()
            except:
                pass
            self.serio = 0
        self.LOCK.free()
        self.Debug("Closed nibeBus serial port: " + self.SERPORT)

    def isOpen(self):
        if self.serio and self.serio.isOpen():
            return 1
        return 0

    def startModule(self):
        if self.LOCK == None:
            self.LOCK = lockfile.LockFile(self.SERPORT)
            
        if not self.isOpen():
            self.openPort()
        if self.isOpen():
            self.start()
            return 1
        else:
            self.setFail()
            self.setTerminated()
            
        return 0


class NibeRS485UDP(NibeRS485Base):
    def __init__(self, listenaddress, port, device):
        NibeRS485Base.__init__(self, device)
       
        self.LISTENADDRESS = listenaddress
        self.UDPPORT = port
        
    def setListenAddress(self, listenaddress):
        self.LISTENADDRESS = listenaddress
                
    def setUDPPort(self, port):
        self.UDPPORT = port
                
    def run(self):        
        sock = None
        try:              
            while self.isRunning():
                try:
                    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                    sock.bind((self.LISTENADDRESS, self.UDPPORT))
                    sock.settimeout(2.0)
                except:
                    try:
                        sock.close()
                    except:
                        pass
                    sock = None
                    self.Log("Cannot create UDP socket for address %s port %d." % (self.LISTENADDRESS, self.UDPPORT))
                    self.setFail()
                    break

                self.Debug("Listening for UDP packets in interface %s port %d" % (self.LISTENADDRESS, self.UDPPORT))

                while self.isRunning():
                    try:
                        (res, peeraddr) = sock.recvfrom(UDP_RECEIVE_BUFFER_SIZE)
                    except socket.timeout, st:
                        continue
                    except Exception, e:
                        self.Log("Socket error. Terminating UDP receiver thread")
                        self.setFail()
                        break
                    
                    self.Debug("Received UDP packet from %s port %d length %d" % (peeraddr[0], peeraddr[1], len(res)))

                    prevchar = ''
                    stat = 1
                    while self.isRunning() and stat != 0 and len(res) > 0:                    
                        (stat, flen) = checkNibeMessage(res, prevchar)
                        if stat > 0:
                            rcvdata = res[:flen]
                            prevchar = rcvdata[flen-1]
                            res = res[flen:]
                            if stat == 1:
                                self.Debug("Received and handled:\n" + log.dumpBuffer(rcvdata))
                                self.handleBuffer(rcvdata)
                            elif stat == 2:
                                self.Debug("Received with invalid CRC:\n" + log.dumpBuffer(rcvdata))
                            else:
                                self.Debug("Received and ignored:\n" + log.dumpBuffer(rcvdata))                                                                                                
                        elif stat == -1:
                            prevchar = res[0]
                            res = res[1:]
                        elif stat == 0:
                            # we cannot get more than there is in the UDP packet, discard the rest
                            self.Log("Discarding %d bytes of data. Incomplete frame." % len(res))
                            res = ''

                sock.close()
                sock = None
                                
                if self.isRunning():
                    time.sleep(SLEEP_AFTER_FAIL)

        except Exception, e:
            self.Log("Exception: " + e.__str__())
            self.setFail()
        except IOError, ioe:
            self.Log("IOError: " + ioe.__str__())
            self.setFail()

        if sock != None:
            try:
                sock.close()
            except:
                pass
            sock = None
       
        self.Log("Nibe Bus UDP receiver thread stopped.")
            
    def startModule(self):            
        self.start()
        return 1

        
###########################################################################

# Functions

# returns (stat, datalen)
#   -1 = error
#   0 = ok, but not ready
#   1 = ok, data lenght in datalen
#   2 = not ok, crc fail, data lenght in datalen
#   3 = ok/not ok, frame not for this device
def checkNibeMessage(data, previousChar = ''):
    l = len(data)
    dl = 0
    if l <= 0:
        return (0, 0)

    # if the previous character is 5C, must abandon one
    # back-to-back 5C can only be in the data part,
    # this is also why CRC is never 5C
    if l >= 1 and previousChar == '\x5C':
        return (-1, 0)

    # first byte is '\x5C'
    if l >= 1 and data[0] != '\x5C':
        return (-1, 0)

    # second byte is '\x00'
    if l >= 2 and data[1] != '\x00':
        return (-1, 0)
    
    # fifth byte is data part len
    if l >= 5:
        dl = ord(data[4])
    else:
        return (0, 0)
    
    # if 1 byte crc and all data have not yet arrived
    if l < dl + 6:
        return (0, 0)
    
    # check third byte address
    # third byte should be '\x20'
    if data[2] != '\x20':
        return (3, dl + 6)

    # check CRC
    crc = 0
    for i in range(2, dl + 5):
        crc = crc ^ ord(data[i])
    datacrc = ord(data[dl + 5])    
    
    if crc == 0x5C:
        if datacrc != 0xC5:
            return (2, dl + 6)
    elif crc != datacrc:
        return (2, dl + 6)

    return (1, dl + 6)

def getNibeDataPart(data):
    dl = ord(data[4])
    return data[5:5+dl]
    
def fixNibeDataPart(data):
    temp = ''
    # remove back_to_back 0x5C characters
    i = 0
    l = len(data)
    while i < l:
        temp = temp + data[i]
        if data[i] == '\x5C' and i+1 < l and data[i+1] == '\x5C':
            i = i + 1
        i = i + 1
    
    return temp
    
def convertNibeMessage(type, value):
    if type == TYPE_INT8:
        temp = binary.twosComplementToInt(value, 8)
        temp = "%d" % temp
        return temp
    elif type == TYPE_INT16:
        temp = binary.twosComplementToInt(value, 16)
        temp = "%d" % temp
        return temp
    elif type == TYPE_INT32:
        temp = binary.twosComplementToInt(value, 32)
        temp = "%d" % temp
        return temp
    elif type == TYPE_UINT8:
        temp = "%d" % value
        return temp
    elif type == TYPE_UINT16:
        temp = "%d" % value
        return temp
    elif type == TYPE_UINT32:
        temp = "%d" % value
        return temp
    elif type == TYPE_INT16_10:
        temp = binary.twosComplementToInt(value, 16)
        temp = "%.1f" % (temp / 10.0)
        return temp
    elif type == TYPE_INT8_10:
        temp = binary.twosComplementToInt(value, 8)
        temp = "%.1f" % (temp / 10.0)
        return temp

    return ""

