#!env python
# -*- coding: iso-8859-1 -*-
###########################################################################
# 
# File:            oumanEH800.py
#
# License:         Donationware, see attached LICENSE file for more 
#                  information
#
# Author:          Olli Lammi (olammi@iki.fi)
#
# Version:         0.1b
#
# Date:            01.12.2014
#
# Description:     Library to communicate with Ouman EH800 controller device
#                  through ethernet interface.
#                  
# Requirements:    Python interpreter 2.4 or newer (www.python.org)
#                  (tested with 2.4.3)
#
# Version history: ** 20.01.2013 v0.1a (Olli Lammi) **
#                  First version. 
#
#                  ** 01.12.2014 v0.1b (Olli Lammi) **
#                  Added ON/OFF-type. 
#
###########################################################################

# Imports

import sys, os 
import string, re
import urllib2

from modules.core import log

###########################################################################

# Constants

POINTTYPE_UNKNOWN = 0
POINTTYPE_TEMP = 1
POINTTYPE_INT = 2
POINTTYPE_STATE = 3


###########################################################################

# Classes

class OumanEH800(log.Logging):
    def __init__(self):
        log.Logging.__init__(self, 'OumanEH800')
        self.URL = ''

    def runQueryCommand(self, cmds):
        dataresult = {}

        datapoints = self.getMeasures(self.URL)
        if len(datapoints.keys()) <= 0:
            self.Log("ERROR: No measurement points in EH800 data or data invalid.")
            return {}
        
        result = self.getRequestData(self.URL, datapoints)
        
        if len(result.keys()) <= 0:
            self.Log("ERROR: No EH800 data values in data or data invalid.")
            return {}
        
        for cmd in cmds:
            if not datapoints.has_key(cmd):
                self.Log("ERROR: Undefined data point: " + cmd)
                continue
            if not result.has_key(cmd):
                self.Log("ERROR: Data key " + cmd + " not found in data result set.")
                continue
            
            type = datapoints[cmd][1]
            value = result[cmd]                
            dataresult[cmd] = handleDataFormat(value, type)
                
        return dataresult
        

    def getMeasures(self, baseurl):
        data = ''
        measuresurl = baseurl + 'measures?;'
        self.Debug("Retrieving Ouman EH800 datapoints from URL: " + measuresurl)
        
        try:
            f = urllib2.urlopen(measuresurl)
            data = f.read()
            f.close()
        except:
            data = ''

        self.Debug("Received measures data: " + repr(data))

        if len(data) <= 0:
            self.Log("Cannot get data from URL: " + measuresurl)
            return {}
        
        return handleMeasures(data)
        

    def getRequestData(self, baseurl, datapoints):
        data = ''
        requesturl = baseurl + 'request?'
        for dp in datapoints.keys():
            requesturl = requesturl + dp + ';'
        self.Debug("Retrieving Ouman EH800 data values from URL: " + requesturl)
        
        try:
            f = urllib2.urlopen(requesturl)
            data = f.read()
            f.close()
        except:
            data = ''

        self.Debug("Received data value data: " + repr(data))

        if len(data) <= 0:
            self.Log("Cannot get data from URL: " + requesturl)
            return {}
        
        return handleRequestData(data)


###########################################################################

# Functions

def handleDataFormat(data, type):
    if type == POINTTYPE_TEMP:
        data = string.strip(data)            
        if len(data) > 0:
            data = "%.1f" % float(data)
    elif type == POINTTYPE_INT:
        data = string.strip(data)
        if len(data) > 0:
            data = "%d" % int(data)
    elif type == POINTTYPE_STATE:
        data = string.strip(data)
        if string.upper(data) == 'ON':
            data = "1"
        elif string.upper(data) == 'OFF':
            data = "0"
 
    return data

def handleMeasures(data):
    res = {}

    data = string.split(data, '\x00')[0]
    if data.startswith('measures?'):
        data = data[9:]
    data = string.split(data, ';')

    for item in data:
        if len(item) <= 0:
            continue
        point = string.split(item, ',')
        if len(point) == 3:
            try:
                res[point[1]] = [point[0], int(point[2])]
            except:
                pass

    return res

def handleRequestData(data):
    res = {}

    data = string.split(data, '\x00')[0]
    if data.startswith('request?'):
        data = data[8:]
    data = string.split(data, ';')

    for item in data:
        if len(item) <= 0:
            continue
        point = string.split(item, '=')
        if len(point) == 2:
            try:
                res[point[0]] = point[1]
            except:
                pass

    return res

