#!env python
# -*- coding: iso-8859-1 -*-
###########################################################################
# 
# File:            oumanSerial.py
#
# License:         Donationware, see attached LICENSE file for more 
#                  information
#
# Author:          Olli Lammi (olammi@iki.fi)
#
# Version:         0.9k
#
# Date:            24.03.2013
#
# Description:     Library to communicate with the Ouman controller using the
#                  Ouman native serial communication protocol. Mainly used to 
#                  read measurement data from the Ouman controller. Will contain
#                  also methods to read and acknowledge alarms and read and alter
#                  controller configuration variables.
#                  
# Supported devices: 
#                  At the moment the supported Ouman EH series controller
#                  devices with serial interface are:
#
#                     EH202S
#                     EH202G
#                     EH203G
#                     EH203GT
#                     EH203T
#                     EH201T  
#                     EH201L 
#                     EH201V
#                     EH201G
#                     EH201GT
#                     EH203
#                     EH686
#                   
#                  If your equipment is not listed, you might still be able to 
#                  use one of the models above to do some logging. You can
#                  also contact the author to add support to a new device or
#                  to improve support for an existing one. 
#
# Requirements:    Python interpreter 2.4 or newer (www.python.org)
#                  (tested with 2.4.3)
# 
#                  Python Serial Port Extension: pySerial  
#                  (http://pyserial.wiki.sourceforge.net/pySerial)
#
# Version history: ** 24.02.2009 v0.9a (Olli Lammi) **
#                  First beta version for testers. 
#
#                  ** 25.02.2009 v0.9c (Olli Lammi) **
#                  Logging and import changes 
#
#                  ** 02.03.2009 v0.9d (Olli Lammi) **
#                  Added locking using a lock file. Changed the module to
#                  open the port only when accessing the serial module.
#
#                  ** 02.06.2009 v0.9e (Olli Lammi) **
#                  Added profile for EH202G (Geopro GS)
#
#                  ** 30.04.2010 v0.9f (Olli Lammi) **
#                  Removed handling of twos complement to 
#                  external module binary.py.
#
#                  ** 18.07.2010 v0.9g (Olli Lammi) **
#                  Added support for writing parameter values. Support for 
#                  EH202S parameters.
#
#                  ** 13.09.2010 v0.9h (Olli Lammi) **
#                  Initial support for EH203GT parameters. 
#
#                  ** 15.11.2010 v0.9i (Olli Lammi) **
#                  Extended the support for EH203GT parameters. 
#
#                  ** 25.01.2013 v0.9j (Olli Lammi) **
#                  Added L2 parameters to EH202S. Added TYPE_VAL32_100
#                  datatype.
#
#                  ** 24.03.2013 v0.9k (Olli Lammi) **
#                  Additions to device profiles.
#
###########################################################################

# Imports

import sys, os 
import string

from modules.core import log
from modules.utils import lockfile
from modules.utils import binary

# pySerial module, not loaded until init of OumanSerial class instance
serial = None

###########################################################################

# Constants

BAUDRATE = 4800 
TIMEOUT = 2

###########################################################################

# Device configuration data

# read/write types
TYPE_READ = 1
TYPE_WRITE = 2 
TYPE_RW = TYPE_READ | TYPE_WRITE

# data types
TYPE_TEMP100 = 1 
TYPE_INT16 = 2
TYPE_UINT16 = 3
TYPE_UINT8 = 4
TYPE_INT8 = 5 
TYPE_UINT32 = 6  
TYPE_INT32 = 7
TYPE_VAL32_100 = 8
TYPE_DATA_LENGTHS = { TYPE_TEMP100: 2, \
                      TYPE_INT16: 2, \
                      TYPE_UINT16: 2, \
                      TYPE_UINT8: 1, \
                      TYPE_INT8: 1, \
                      TYPE_UINT32: 4, \
                      TYPE_INT32: 4, \
                      TYPE_VAL32_100: 4 }

# device_id: data_point_name, id, datatype, readwritetype, offset_in_bytes
OUMAN_DEVICES = { 'EH202S': [ ['Ulkolampotila', 18, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1 huone', 21, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1 menovesi', 20, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L2 menovesi', 26, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Vapaa mittaus', 27, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Varaaja ylatulo', 24, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Varaaja yla', 33, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Varaaja keski', 23, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Varaaja ala', 34, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Maaliuos', 25, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1-venttiili', 49, TYPE_UINT8, TYPE_READ, 0], \
                              ['L2-venttiili', 50, TYPE_UINT8, TYPE_READ, 0], \
                              ['JV-venttiili', 51, TYPE_UINT8, TYPE_READ, 0], \
                              ['SV tunnit', 58, TYPE_INT16, TYPE_READ, 0], \
                              ['Kompressori tunnit', 59, TYPE_INT16, TYPE_READ, 0], \
                              ['Tilamuuttuja', 45, TYPE_UINT16, TYPE_READ, 0], \
                              ['L1 -20', 0x0C, TYPE_INT8, TYPE_RW, 0], \
                              ['L1 0', 0x0C, TYPE_INT8, TYPE_RW, 1], \
                              ['L1 20', 0x0C, TYPE_INT8, TYPE_RW, 2], \
                              ['L1 Yoalennus vaikutus', 0x0C, TYPE_INT8, TYPE_RW, 5], \
                              ['L1 Minimiraja', 0x0C, TYPE_INT8, TYPE_RW, 3], \
                              ['L1 Maksimiraja', 0x0C, TYPE_INT8, TYPE_RW, 4], \
                              ['L1 Syyskuivaus', 0x0C, TYPE_INT8, TYPE_RW, 14], \
                              ['L1 Ulkolampotilan hidastus', 0x0C, TYPE_INT8, TYPE_RW, 19], \
                              ['L1 Esikorotus', 0x0C, TYPE_INT8, TYPE_RW, 23], \
                              ['L1 Venttiilin kesasulkeutuminen', 0x2C, TYPE_INT8, TYPE_RW, 8], \
                              ['L2 -20', 0x0E, TYPE_INT8, TYPE_RW, 0], \
                              ['L2 0', 0x0E, TYPE_INT8, TYPE_RW, 1], \
                              ['L2 20', 0x0E, TYPE_INT8, TYPE_RW, 2], \
                              ['L2 Yoalennus vaikutus', 0x0E, TYPE_INT8, TYPE_RW, 5], \
                              ['L2 Minimiraja', 0x0E, TYPE_INT8, TYPE_RW, 3], \
                              ['L2 Maksimiraja', 0x0E, TYPE_INT8, TYPE_RW, 4], \
                              ['L2 Syyskuivaus', 0x0E, TYPE_INT8, TYPE_RW, 9], \
                              ['L2 Ulkolampotilan hidastus', 0x0E, TYPE_INT8, TYPE_RW, 15], \
                              ['L2 Esikorotus', 0x0E, TYPE_INT8, TYPE_RW, 19], \
                              ['ML Ohjaustapa', 0x0B, TYPE_INT8, TYPE_RW, 10], \
                              ['ML VaraajaAlaMin', 0x0B, TYPE_INT8, TYPE_RW, 5], \
                              ['ML SV_viiveaika', 0x0B, TYPE_INT8, TYPE_RW, 6], \
                              ['ML VarYla Kesa', 0x0B, TYPE_INT8, TYPE_RW, 2], \
                              ['ML VarYla Talvi', 0x0B, TYPE_INT8, TYPE_RW, 1], \
                              ['ML Lisalammitys', 0x0B, TYPE_INT8, TYPE_RW, 8], \
                              ['ML Lisalam aika', 0x0B, TYPE_INT8, TYPE_RW, 9], \
                              ['ML Kompr esto', 0x0B, TYPE_INT8, TYPE_RW, 14], \
                              ['ML VarYlaEroSV', 0x0B, TYPE_INT8, TYPE_RW, 7], \
                              ['ML VaraajaAlaEro', 0x0B, TYPE_INT8, TYPE_RW, 3], \
                              ['ML VaraajaYlaEro', 0x0B, TYPE_INT8, TYPE_RW, 4], \
                              ['ML JV esiavaus', 0x0B, TYPE_INT8, TYPE_RW, 12], \
                              ['ML JV Esto', 0x0B, TYPE_UINT8, TYPE_RW, 13], \
                              ['ML Maaliuos halytys', 0x0B, TYPE_INT8, TYPE_RW, 11] \
                            ], \
                  'EH202G': [ ['Ulkolampotila', 18, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1 huone', 21, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1 menovesi', 20, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L2 menovesi', 26, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Vapaa mittaus', 27, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Varaaja ala', 24, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Varaaja yla', 33, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1 paluuvesi', 23, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Lammitys varaaja ala', 34, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Maaliuos', 25, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Tilamuuttuja', 45, TYPE_UINT16, TYPE_READ, 0] \
                            ], \
                  'EH203G': [ ['Ulkolampotila', 18, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1 huone', 21, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1 menovesi', 20, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L2 menovesi', 26, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Vapaa mittaus', 27, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Varaaja yla', 33, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Varaaja ala', 34, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Maaliuos', 25, TYPE_TEMP100, TYPE_READ, 0], \
                              ['LV menovesi', 24, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Kuumakaasu', 41, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1-venttiili', 49, TYPE_UINT8, TYPE_READ, 0], \
                              ['L2-venttiili', 50, TYPE_UINT8, TYPE_READ, 0], \
                              ['Tilamuuttuja', 45, TYPE_UINT16, TYPE_READ, 0] \
                            ], \
                  'EH203GT': [ ['Ulkolampotila', 18, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1 huone', 21, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1 menovesi', 20, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L2 menovesi', 26, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Vapaa mittaus', 27, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Varaaja yla', 33, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Varaaja ala', 34, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Maaliuos', 25, TYPE_TEMP100, TYPE_READ, 0], \
                              ['LV menovesi', 24, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Kuumakaasu', 41, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1-venttiili', 49, TYPE_UINT8, TYPE_READ, 0], \
                              ['SV tunnit', 58, TYPE_INT16, TYPE_READ, 0], \
                              ['Kompressori tunnit', 59, TYPE_INT16, TYPE_READ, 0], \
                              ['Tilamuuttuja', 45, TYPE_UINT16, TYPE_READ, 0], \
                              ['L1 -20', 0x0C, TYPE_INT8, TYPE_RW, 0], \
                              ['L1 0', 0x0C, TYPE_INT8, TYPE_RW, 1], \
                              ['L1 20', 0x0C, TYPE_INT8, TYPE_RW, 2], \
                              ['L1 Yoalennus vaikutus', 0x0C, TYPE_INT8, TYPE_RW, 5], \
                              ['L1 Minimiraja', 0x0C, TYPE_INT8, TYPE_RW, 3], \
                              ['L1 Maksimiraja', 0x0C, TYPE_INT8, TYPE_RW, 4], \
                              ['L1 Syyskuivaus', 0x0C, TYPE_INT8, TYPE_RW, 14], \
                              ['L1 Ulkolampotilan hidastus', 0x0C, TYPE_INT8, TYPE_RW, 19], \
                              ['L1 Esikorotus', 0x0C, TYPE_INT8, TYPE_RW, 6], \
                              ['L1 Venttiilin kesasulkeutuminen', 0x2C, TYPE_INT8, TYPE_RW, 8], \
                              ['L1 MenovesiMin 0', 0x0C, TYPE_INT8, TYPE_RW, 8], \
                              ['L1 MenovesiMin -20', 0x0C, TYPE_INT8, TYPE_RW, 9], \
                              ['L1 PoikkeamaHal', 0x0C, TYPE_INT8, TYPE_RW, 26], \
                              ['L1 PoikkeamaHalHidastus', 0x2C, TYPE_INT8, TYPE_RW, 28], \
                              ['L1 Ohjaustapa', 0x2A, TYPE_INT8, TYPE_RW, 1], \
                              ['L2 -20', 0x0E, TYPE_INT8, TYPE_RW, 0], \
                              ['L2 0', 0x0E, TYPE_INT8, TYPE_RW, 1], \
                              ['L2 20', 0x0E, TYPE_INT8, TYPE_RW, 2], \
                              ['L2 Yoalennus vaikutus', 0x0C, TYPE_INT8, TYPE_RW, 5], \
                              ['L2 Minimiraja', 0x0E, TYPE_INT8, TYPE_RW, 3], \
                              ['L2 Maksimiraja', 0x0E, TYPE_INT8, TYPE_RW, 4], \
                              ['L2 Syyskuivaus', 0x0E, TYPE_INT8, TYPE_RW, 9], \
                              ['L2 Ulkolampotilan hidastus', 0x0E, TYPE_INT8, TYPE_RW, 15], \
                              ['L2 Esikorotus', 0x0E, TYPE_INT8, TYPE_RW, 19], \
                              ['L2 MenovesiMin 0', 0x0E, TYPE_INT8, TYPE_RW, 7], \
                              ['L2 MenovesiMin -20', 0x0E, TYPE_INT8, TYPE_RW, 8], \
                              ['L2 PoikkeamaHal', 0x0E, TYPE_INT8, TYPE_RW, 22], \
                              ['LV Kayttovesi', 0x10, TYPE_INT8, TYPE_RW, 2], \
                              ['LV Halytys', 0x10, TYPE_INT8, TYPE_RW, 6], \
                              ['LV PoikkeamaHal', 0x10, TYPE_INT8, TYPE_RW, 24], \
                              ['LV PoikkeamaHalHidastus', 0x10, TYPE_INT8, TYPE_RW, 22], \
                              ['LV Ohjaustapa', 0x2A, TYPE_INT8, TYPE_RW, 0], \
                              ['ML TaysitehoVarYlaMin', 0x0B, TYPE_INT8, TYPE_RW, 0], \
                              ['ML OsatehoVarYlaMin', 0x0B, TYPE_INT8, TYPE_RW, 1], \
                              ['ML YlaEro', 0x0B, TYPE_INT8, TYPE_RW, 2], \
                              ['ML TaysitehoYlaEroSV', 0x0B, TYPE_INT8, TYPE_RW, 3], \
                              ['ML OsatehoYlaEroSV', 0x0B, TYPE_INT8, TYPE_RW, 4], \
                              ['ML VaraajaAlaMin', 0x0B, TYPE_INT8, TYPE_RW, 5], \
                              ['ML VaraajaAlaEro', 0x0B, TYPE_INT8, TYPE_RW, 6], \
                              ['ML OsatehoVarYlaKomp', 0x0B, TYPE_INT8, TYPE_RW, 7] \
                            ], \
                  'EH203T': [ ['Ulkolampotila', 18, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1 huone', 21, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1 menovesi', 20, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Vapaa mittaus', 27, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Varaaja yla', 33, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Varaaja ala', 34, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Maaliuos', 25, TYPE_TEMP100, TYPE_READ, 0], \
                              ['LV menovesi', 24, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Kuumakaasu', 41, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1-venttiili', 49, TYPE_UINT8, TYPE_READ, 0], \
                              ['Tilamuuttuja', 45, TYPE_UINT16, TYPE_READ, 0] \
                            ], \
                  'EH201T': [ ['Ulkolampotila', 18, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1 huone', 21, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1 menovesi', 20, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Varaaja yla', 33, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Varaaja ala', 34, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Kuumakaasu', 41, TYPE_TEMP100, TYPE_READ, 0], \
                              ['SV tunnit', 58, TYPE_INT16, TYPE_READ, 0], \
                              ['Kompressori tunnit', 59, TYPE_INT16, TYPE_READ, 0], \
                              ['Tilamuuttuja', 45, TYPE_UINT16, TYPE_READ, 0] \
                            ], \
                  'EH201L': [ ['Ulkolampotila', 18, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1 huone', 21, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1 menovesi', 20, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1 paluuvesi', 23, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Mittaus 9', 33, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Mittaus 10', 34, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Mittaus 11', 41, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1-venttiili', 49, TYPE_UINT8, TYPE_READ, 0], \
                              ['Tilamuuttuja', 45, TYPE_UINT16, TYPE_READ, 0] \
                            ], \
                  'EH201V': [ ['LV menovesi', 24, TYPE_TEMP100, TYPE_READ, 0], \
                              ['LV kiertovesi', 25, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Mittaus 9', 33, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Mittaus 10', 34, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Mittaus 11', 41, TYPE_TEMP100, TYPE_READ, 0], \
                              ['LV-venttiili', 51, TYPE_UINT8, TYPE_READ, 0], \
                              ['Tilamuuttuja', 45, TYPE_UINT16, TYPE_READ, 0] \
                            ], \
                  'EH201G': [ ['Ulkolampotila', 18, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1 huone', 21, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1 menovesi', 20, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Vapaa mittaus', 27, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Varaaja yla', 33, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Varaaja ala', 34, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Kuumakaasu', 41, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1-venttiili', 49, TYPE_UINT8, TYPE_READ, 0], \
                              ['Tilamuuttuja', 45, TYPE_UINT16, TYPE_READ, 0] \
                            ], \
                  'EH201GT': [ ['Ulkolampotila', 18, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1 huone', 21, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1 menovesi', 20, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Varaaja yla', 33, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Varaaja ala', 34, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Kuumakaasu', 41, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Tilamuuttuja', 45, TYPE_UINT16, TYPE_READ, 0], \
                              ['L1 -20', 0x0C, TYPE_INT8, TYPE_RW, 0], \
                              ['L1 0', 0x0C, TYPE_INT8, TYPE_RW, 1], \
                              ['L1 20', 0x0C, TYPE_INT8, TYPE_RW, 2], \
                              ['L1 Yoalennus vaikutus', 0x0C, TYPE_INT8, TYPE_RW, 5], \
                              ['L1 Minimiraja', 0x0C, TYPE_INT8, TYPE_RW, 3], \
                              ['L1 Maksimiraja', 0x0C, TYPE_INT8, TYPE_RW, 4], \
                              ['L1 Syyskuivaus', 0x0C, TYPE_INT8, TYPE_RW, 14], \
                              ['L1 Ulkolampotilan hidastus', 0x0C, TYPE_INT8, TYPE_RW, 19], \
                              ['L1 Esikorotus', 0x0C, TYPE_INT8, TYPE_RW, 6], \
                              ['L1 Venttiilin kesasulkeutuminen', 0x2C, TYPE_INT8, TYPE_RW, 8], \
                              ['L1 MenovesiMin 0', 0x0C, TYPE_INT8, TYPE_RW, 8], \
                              ['L1 MenovesiMin -20', 0x0C, TYPE_INT8, TYPE_RW, 9], \
                              ['L1 PoikkeamaHal', 0x0C, TYPE_INT8, TYPE_RW, 26], \
                              ['L1 PoikkeamaHalHidastus', 0x2C, TYPE_INT8, TYPE_RW, 28], \
                              ['L1 Ohjaustapa', 0x2A, TYPE_INT8, TYPE_RW, 1] \
                            ], \
                  'EH203': [ ['Ulkolampotila', 18, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1 huone', 21, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1 menovesi', 20, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L1 paluuvesi', 23, TYPE_TEMP100, TYPE_READ, 0], \
                              ['L2 menovesi', 26, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Mittaus 6', 27, TYPE_TEMP100, TYPE_READ, 0], \
                              ['LV menovesi', 24, TYPE_TEMP100, TYPE_READ, 0], \
                              ['LV kiertovesi', 25, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Mittaus 9', 33, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Mittaus 10', 34, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Mittaus 11', 41, TYPE_TEMP100, TYPE_READ, 0], \
                              ['Tilamuuttuja', 45, TYPE_UINT16, TYPE_READ, 0], \
                              ['L1-venttiili', 49, TYPE_UINT8, TYPE_READ, 0], \
                              ['L2-venttiili', 50, TYPE_UINT8, TYPE_READ, 0], \
                              ['LV-venttiili', 51, TYPE_UINT8, TYPE_READ, 0], \
                              ['Huipputeho', 60, TYPE_INT32, TYPE_READ, 0], \
                              ['Huippuvirtaus', 61, TYPE_VAL32_100, TYPE_READ, 0], \
                              ['KL energia', 63, TYPE_INT32, TYPE_READ, 0], \
                              ['KL vesi', 64, TYPE_VAL32_100, TYPE_READ, 0] \
                            ], \
                  'EH686': [ ['Tilamuuttuja', 45, TYPE_UINT16, TYPE_READ, 0], \
                             ['Kuittaus', 55, TYPE_UINT8, TYPE_RW, 0], \
                             ['Ulkolampotila', 18, TYPE_TEMP100, TYPE_READ, 0], \
                             ['Huonelampotila', 21, TYPE_TEMP100, TYPE_READ, 0], \
                             ['M1', 27, TYPE_VAL32_100, TYPE_READ,0], \
                             ['M1tila', 27, TYPE_INT32, TYPE_READ,0], \
                             ['M2', 28, TYPE_VAL32_100, TYPE_READ,0], \
                             ['M2tila', 28, TYPE_INT32, TYPE_READ,0], \
                             ['M3', 29, TYPE_VAL32_100, TYPE_READ,0], \
                             ['M3tila', 29, TYPE_INT32, TYPE_READ,0], \
                             ['M4', 30, TYPE_VAL32_100, TYPE_READ,0], \
                             ['M4tila', 30, TYPE_INT32, TYPE_READ,0], \
                             ['M5', 31, TYPE_VAL32_100, TYPE_READ,0], \
                             ['M5tila', 31, TYPE_INT32, TYPE_READ,0], \
                             ['M6', 32, TYPE_VAL32_100, TYPE_READ,0], \
                             ['M6tila', 32, TYPE_INT32, TYPE_READ,0], \
                             ['M7', 33, TYPE_VAL32_100, TYPE_READ,0], \
                             ['M7tila', 33, TYPE_INT32, TYPE_READ,0], \
                             ['M8', 34, TYPE_VAL32_100, TYPE_READ,0], \
                             ['M8tila', 34, TYPE_INT32, TYPE_READ,0], \
                           ], \
                  'DEBUG': [ ['Kuittaus', 55, TYPE_UINT8, TYPE_RW, 0] ] \
                }

TILA_MASK_DIGI1 = 1 
TILA_MASK_DIGI2 = 2 
TILA_MASK_DIGI3 = 4 
TILA_MASK_SAHKOVASTUS = 8 
TILA_MASK_KOMPRESSORI = 16

TILA_ML_OHJAUSTAPA_KOMPR_SV = 0
TILA_ML_OHJAUSTAPA_KOMPR = 1
TILA_ML_OHJAUSTAPA_SV = 2


###########################################################################

# Classes

class OumanSerial(log.Logging):
    def __init__(self, port, device):
        log.Logging.__init__(self, 'OumanSerial')
        
        global serial
        serial = __import__('serial') 
        
        self.SERPORT = port
        self.OUMAN_DEVICE = device
        self.serio = 0
          
        self.LOCK = None
    
    def setPort(self, port):
        self.SERPORT = port
        
    def setDevice(self, device):
        self.OUMAN_DEVICE = device
    
    def openPort(self):
        if not OUMAN_DEVICES.has_key(self.OUMAN_DEVICE):
            self.Log("ERROR: Invalid OUMAN device type: " + self.OUMAN_DEVICE)
            return 0
        
        if self.LOCK == None:
            self.LOCK = lockfile.LockFile(self.SERPORT)

        if not self.LOCK.lock():
            self.Log("ERROR: Unable to aquire lockfile for oumanSerial port: " + self.LOCK.getName())
            return 0

        try:
            self.serio = serial.Serial(port=self.SERPORT,baudrate=BAUDRATE, \
                                bytesize=serial.EIGHTBITS, \
                                parity=serial.PARITY_NONE, stopbits=serial.STOPBITS_ONE, \
                                xonxoff=0, rtscts=0, dsrdtr=0, timeout=TIMEOUT)
            if not self.serio.isOpen():
                self.serio = 0
                self.LOCK.free()
                return 0 
        except:
            self.serio = 0
            self.LOCK.free()
            return 0
        
        self.Debug("Opened oumanSerial port: " + self.SERPORT)
        return 1
        
    def closePort(self):
        if self.serio:
            try:
                self.serio.close()
            except:
                pass
            self.serio = 0
        self.LOCK.free()
        self.Debug("Closed oumanSerial port: " + self.SERPORT)

    def isOpen(self):
        if self.serio and self.serio.isOpen():
            return 1
        return 0

    def runQueryCommand(self, cmd):
        if not OUMAN_DEVICES.has_key(self.OUMAN_DEVICE):
            self.Log("ERROR: Invalid OUMAN device type: " + self.OUMAN_DEVICE)
            return ""
        
        id = 0
        type = 0
        offset = 0
        for t in OUMAN_DEVICES[self.OUMAN_DEVICE]:
            if t[0] == cmd:
                id = t[1]
                type = t[2]
                offset = t[4]
                break

        if type == 0:
            self.Log("ERROR: Invalid OUMAN key: " + cmd)
            return ""

        res = self.runQueryId(id)
        if len(res) > 0:
            return convertOumanMessage(type, offset, res)
        return ""

    def runSetCommand(self, cmd, value):
        if not OUMAN_DEVICES.has_key(self.OUMAN_DEVICE):
            self.Log("ERROR: Invalid OUMAN device type: " + self.OUMAN_DEVICE)
            return ""
        
        # check if the parameter is writable
        id = 0
        type = 0
        wrtype = 0
        offset = 0
        for t in OUMAN_DEVICES[self.OUMAN_DEVICE]:
            if t[0] == cmd:
                id = t[1]
                type = t[2]
                wrtype = t[3]
                offset = t[4]
                break

        if type == 0:
            self.Log("ERROR: Invalid OUMAN key: " + cmd)
            return ""

        if wrtype & TYPE_WRITE == 0:
            self.Log("ERROR: OUMAN parameter " + cmd + " is not writable.")
            return ""

        data = convertToOumanData(type, value)
        if len(data) <= 0:
            self.Log("ERROR: Invalid data value for parameter type.")
            return ""
            
        res = self.runQueryId(id)
        if len(res) <= 0:
            self.Log("ERROR: Cannot get data from Ouman controller.")
            return ""

        # alter data with new value to the correct offset        
        tlen = TYPE_DATA_LENGTHS[type] 
        olddata = getOumanDataPart(res)[2:]
        newdata = olddata[:offset] + data[:tlen] + olddata[offset+tlen:]
         
        # write back to Ouman
        resdata = self.runWriteId(id, newdata)
        
        if len(resdata) <= 0:
            self.Log("ERROR: Cannot get response data from Ouman controller.")
            return ""
            
        # check that the response contains correct id
        resid = getOumanDataPart(resdata)
        if len(resid) < 2 or (ord(resid[0]) << 8) + ord(resid[1]) != id:
            self.Log("ERROR: Invalid response data from Ouman controller.")
            return ""
                    
        return "OK"

    def runQueryId(self, id):
        self.Debug("Running oumanSerial query with id " + repr(id))
        if not self.isOpen():
            if not self.openPort():
                self.Log("ERROR: Cannot open serial port.")
                return ""
            (stat, res) = serialQuery(self.serio, id)
            self.closePort()
        else:
            (stat, res) = serialQuery(self.serio, id)
        self.Debug("Received ouman response: " + repr(res))
        
        if stat == 1:
            return res
        elif stat == -1:
            self.Log("ERROR: Timeout reading message from Ouman")
            return ""
        elif stat == -2:
            self.Log("ERROR: Error in return message from Ouman")
            return ""

    def runWriteId(self, id, data):
        self.Debug("Running oumanSerial write with id " + repr(id))
        if not self.isOpen():
            if not self.openPort():
                self.Log("ERROR: Cannot open serial port.")
                return ""
            (stat, res) = serialWrite(self.serio, id, data)
            self.closePort()
        else:
            (stat, res) = serialWrite(self.serio, id, data)
        self.Debug("Received ouman response: " + repr(res))
        
        if stat == 1:
            return res
        elif stat == -1:
            self.Log("ERROR: Timeout reading message from Ouman")
            return ""
        elif stat == -2:
            self.Log("ERROR: Error in return message from Ouman")
            return ""


###########################################################################

# Functions

# returns (status, response)
# status: 1 = ok, -1 = timeout, -2 = error in response
def serialQuery(serio, id):
    crc = (0x81 + 0x02 + 0x00 + id) & 0xFF
    sendbuff = '\x02\x81\x02\x00' + chr(id) + chr(crc)

    serio.write(sendbuff)
    serio.flush()

    res = ""
    temp = " "
    while len(temp) > 0:
        temp = serio.read()
        if len(temp) <= 0:
            if checkOumanMessage(res) == 1:
                return (1, res)
            else:
                return (-1, '')
        res = res + temp
        stat = checkOumanMessage(res)
        if stat == 1:
            return (1, res)
        elif stat == 2:
            return (-2, '')


# returns (status, response)
# status: 1 = ok, -1 = timeout, -2 = error in response
def serialWrite(serio, id, data):
    datacrc = 0
    for i in range(len(data)):
      datacrc = datacrc + ord(data[i])
    crc = (0x82 + 0x02 + len(data) + 0x00 + id + datacrc) & 0xFF
    sendbuff = '\x02\x82' + chr(2 + len(data)) + '\x00' + chr(id) + data + chr(crc)

    serio.write(sendbuff)
    serio.flush()

    res = ""
    temp = " "
    while len(temp) > 0:
        temp = serio.read()
        if len(temp) <= 0:
            if checkOumanMessage(res) == 1:
                return (1, res)
            else:
                return (-1, '')
        res = res + temp
        stat = checkOumanMessage(res)
        if stat == 1:
            return (1, res)
        elif stat == 2:
            return (-2, '')

# returns int
# 0 = ok, but not ready
# 1 = ok
# 2 = error
def checkOumanMessage(data):
    if len(data) <= 0:
        return 0

    # first byte is 0x02
    if data[0] != chr(0x02):
        return 2

    # ignore second byte
    if len(data) < 3:
        return 0

    # byte 3 is content lenght
    dlen = ord(data[2])

    # if package is already longer than it should
    if len(data) > dlen + 4:
        return 2

    if len(data) < dlen + 4:
        return 0

    # if package crc does not match
    crc = 0
    for i in range(1, len(data)-1):
        crc = crc + ord(data[i])
    if (crc & 0xFF) != ord(data[len(data)-1]):
        return 2

    return 1


def getOumanDataPart(data):
    if checkOumanMessage(data) != 1:
        return ""

    dlen = ord(data[2])
    return data[3: 3+dlen]


def convertOumanMessage(type, offset, data):
    # get data part and remove id part
    data = getOumanDataPart(data)

    if len(data) <= 2:
        return ""
    data = data[2:]
    tlen = TYPE_DATA_LENGTHS[type]

    if offset > 0:
        if len(data) < (offset + tlen): 
            return ""
        data = data[offset:]

    if type == TYPE_TEMP100:
        temp = 0
        temp = (ord(data[0]) << 8) + ord(data[1])
        temp = binary.twosComplementToInt(temp, 16)
        temp = "%.2f" % (temp / 100.0)
        return temp
    elif type == TYPE_INT16:
        temp = (ord(data[0]) << 8) + ord(data[1])
        temp = binary.twosComplementToInt(temp, 16)
        temp = "%d" % temp
        return temp
    elif type == TYPE_UINT16:
        temp = (ord(data[0]) << 8) + ord(data[1])
        temp = "%d" % temp
        return temp
    elif type == TYPE_INT8:
        temp = ord(data[0])
        temp = binary.twosComplementToInt(temp, 8)
        temp = "%d" % temp
        return temp
    elif type == TYPE_UINT8:
        temp = ord(data[0])
        temp = "%d" % temp
        return temp
    elif type == TYPE_UINT32:
        temp = (ord(data[0]) << 24) + (ord(data[1]) << 16)
        temp = temp + (ord(data[2]) << 8) + ord(data[3])
        temp = "%d" % temp
        return temp
    elif type == TYPE_INT32:
        temp = (ord(data[0]) << 24) + (ord(data[1]) << 16)
        temp = temp + (ord(data[2]) << 8) + ord(data[3])
        temp = binary.twosComplementToInt(temp, 32)
        temp = "%d" % temp
        return temp
    elif type == TYPE_VAL32_100:
        temp = (ord(data[0]) << 24) + (ord(data[1]) << 16)
        temp = temp + (ord(data[2]) << 8) + ord(data[3])
        temp = binary.twosComplementToInt(temp, 32)
        temp = "%.2f" % (temp / 100.0)
        return temp

    return ""


def convertToOumanData(type, data):
    if len(data) <= 0:
        return ""
    tlen = TYPE_DATA_LENGTHS[type]

    if type == TYPE_TEMP100:
        temp = int(float(data) * 100.0)
        temp = binary.intToTwosComplement(temp, 16)
        return chr((temp >> 8) & 0xFF) + chr(temp & 0xFF) 
    elif type == TYPE_INT16:
        temp = int(data)
        temp = binary.intToTwosComplement(temp, 16)
        return chr((temp >> 8) & 0xFF) + chr(temp & 0xFF) 
    elif type == TYPE_UINT16:
        temp = int(data)
        return chr((temp >> 8) & 0xFF) + chr(temp & 0xFF) 
    elif type == TYPE_INT8:
        temp = int(data)
        temp = binary.intToTwosComplement(temp, 8)
        return chr(temp & 0xFF) 
    elif type == TYPE_UINT8:
        temp = int(data)
        return chr(temp & 0xFF) 
    elif type == TYPE_UINT32:
        temp = int(data)
        return chr((temp >> 24) & 0x00FF) + chr((temp >> 16) & 0x00FF) + \
               chr((temp >> 8) & 0x00FF) + chr(temp & 0x00FF) 
    elif type == TYPE_INT32:
        temp = int(data)
        temp = binary.intToTwosComplement(temp, 32)
        return chr((temp >> 24) & 0x00FF) + chr((temp >> 16) & 0x00FF) + \
               chr((temp >> 8) & 0x00FF) + chr(temp & 0x00FF) 
    elif type == TYPE_VAL32_100:
        temp = int(float(data) * 100.0)
        temp = binary.intToTwosComplement(temp, 32)
        return chr((temp >> 24) & 0x00FF) + chr((temp >> 16) & 0x00FF) + \
               chr((temp >> 8) & 0x00FF) + chr(temp & 0x00FF) 
    
    return ""

            
