#!env python
# -*- coding: iso-8859-1 -*-
###########################################################################
# 
# File:            regoCanSerial.py
#
# License:         Donationware, see attached LICENSE file for more 
#                  information
#
# Author:          Olli Lammi (olammi@iki.fi)
#                  (initial data provided by Raimo Ruokonen) 
#
# Version:         0.1b
#
# Date:            13.02.2013
#
# Description:     Library to communicate with the Rego Can bus type heat 
#                  pump controllers.
#                  
# Supported devices: 
#                  Rego 800 and Rego 1000 equipped devices. 
#                  Tested with:
#                      REGO1000: IVT EQ8
#                      REGO800:  - 
#
# Requirements:    Python interpreter 2.4 or newer (www.python.org)
#                  (tested with 2.4.3)
# 
#                  Python Serial Port Extension: pySerial  
#                  (http://pyserial.wiki.sourceforge.net/pySerial)
#
# Version history: ** 26.11.2012 v0.1a (Olli Lammi) **
#                  First test version. 
#
#                  ** 13.02.2013 v0.1b (Olli Lammi) **
#                  Corrected typo and added description "Heat fluid pump 
#                  control" to REGO1000 descriptions. 
#
###########################################################################

# Imports

import sys, os 
import string, math
import tempfile
import time 

from modules.core import threads
from modules.core import log
from modules.utils import lockfile
from modules.utils import binary
from modules.utils import can232Serial


###########################################################################

# Constants

DATAVALID = 240

###########################################################################

# Device configuration data

TYPE_TEMP16 = 1
TYPE_UINT8 = 2
TYPE_UINT16 = 3

REGOCAN_DEVICES = { 'REGO1000': [ 125, [ \
                                 [0x08000270, 'Heat carrier 1', TYPE_TEMP16], \
                                 [0x08004270, 'Outdoor', TYPE_TEMP16], \
                                 [0x08008270, 'Warm water', TYPE_TEMP16], \
                                 [0x0800c270, 'Room 1', TYPE_TEMP16], \
                                 [0x08010270, 'Hot gas', TYPE_TEMP16], \
                                 [0x08014270, 'Heat fluid out', TYPE_TEMP16], \
                                 [0x08018270, 'Heat fluid in', TYPE_TEMP16], \
                                 [0x0801c270, 'Cold fluid in', TYPE_TEMP16], \
                                 [0x08020270, 'Cold fluid out', TYPE_TEMP16], \
                                 [0x00028270, 'Heat carrier pump 1', TYPE_UINT8], \
                                 [0x0002c270, 'Heat fluid pump', TYPE_UINT8], \
                                 [0x00030270, 'xxx0030', TYPE_UINT8], \
                                 [0x00034270, 'xxx0034', TYPE_UINT8], \
                                 [0x00038270, 'Three-way valve', TYPE_UINT8], \
                                 [0x0003c270, 'Additional heat 1', TYPE_UINT8], \
                                 [0x00040270, 'xxx0040', TYPE_UINT8], \
                                 [0x00044270, 'xxx0044', TYPE_UINT8], \
                                 [0x00048270, 'Compressor', TYPE_UINT8], \
                                 [0x0004c270, 'Additional heat 2', TYPE_UINT8], \
                                 [0x00050270, 'xxx0050', TYPE_UINT8], \
                                 [0x00054270, 'Cold fluid pump', TYPE_UINT8], \
                                 [0x00090270, 'Heat fluid pump control', TYPE_UINT8], \
                                 [0x08058270, 'xxx8058', TYPE_UINT8], \
                                 [0x0805c270, 'xxx805c', TYPE_UINT8], \
                                 [0x08060270, 'xxx8060', TYPE_UINT8], \
                                 [0x08064270, 'xxx8064', TYPE_UINT8], \
                                 [0x08068270, 'xxx8068', TYPE_UINT8], \
                                 [0x0806c270, 'xxx806c', TYPE_UINT8], \
                                 [0x08070270, 'xxx8070', TYPE_UINT8], \
                                 [0x08074270, 'xxx8074', TYPE_UINT8], \
                                 [0x08078270, 'xxx8078', TYPE_UINT8], \
                                 [0x0807c270, 'xxx807c', TYPE_UINT8] \
                               ] ], \
                    'REGO800': [ 125, [ \
                                 [0x08000260, 'Heat carrier 1', TYPE_TEMP16], \
                                 [0x08004260, 'Outdoor', TYPE_TEMP16], \
                                 [0x08008260, 'Warm water', TYPE_TEMP16], \
                                 [0x0800c260, 'xxx800c', TYPE_TEMP16], \
                                 [0x08010260, 'Hot gas', TYPE_TEMP16], \
                                 [0x08014260, 'Heat fluid out', TYPE_TEMP16], \
                                 [0x08018260, 'Heat fluid in', TYPE_TEMP16], \
                                 [0x0801c260, 'Cold fluid in', TYPE_TEMP16], \
                                 [0x08020260, 'Cold fluid out', TYPE_TEMP16] \
                               ] ] \
                  }

   

###########################################################################

# Classes

class RegoCanSerial(log.Logging, can232Serial.CANFrameListener):
    def __init__(self, port, device, baudrate):
        log.Logging.__init__(self, 'RegoCanSerial')
               
        self.data = {}
        self.data_lock = threads.Lock()
  
        self.SERPORT = port
        self.REGOCAN_DEVICE = device
        self.BAUDRATE = baudrate
        self.can232 = None
          
    def setPort(self, port):
        self.SERPORT = port
        
    def setDevice(self, device):
        self.REGOCAN_DEVICE = device
        
    def setBaudrate(self, baudrate):
        self.BAUDRATE = baudrate      
                    
    def frameReceived(self, id, data):
        if self.data_lock.lock_wait():
            self.data[id] = [time.time(), data]
            self.Debug("Got data for id %d/0x%X, data lenght %d: %s" % (id, id, len(data), repr(data)))
            self.data_lock.free()
        else:
            self.Log("Cannot get data lock.")

    def startModule(self):
        if not self.isOpen():
            self.openPort()
        if self.isOpen():
            return 1

        return 0
    
    def terminate(self):
        if self.isOpen():
            self.closePort()
        return 1
                   
    def openPort(self):
        if not REGOCAN_DEVICES.has_key(self.REGOCAN_DEVICE):
            self.Log("ERROR: Invalid REGOCAN device type: " + self.REGOCAN_DEVICE)
            return 0
        canspeedkbps = REGOCAN_DEVICES[self.REGOCAN_DEVICE][0]

        self.can232 = can232Serial.Can232Serial(self.SERPORT, canspeedkbps, self.BAUDRATE)
        self.can232.setFrameListener(self)
        if not self.can232.openPort():
            self.Debug("Cannot open regoCanSerial port: " + self.SERPORT)
            self.can232 = None
            return 0
        
        self.Debug("Opened regoCanSerial port: " + self.SERPORT)
        return 1
        
    def closePort(self):
        if self.can232:
            try:
                self.can232.closePort()
            except:
                pass
            self.can232 = None
        self.Debug("Closed regoCanSerial port: " + self.SERPORT)

    def isOpen(self):
        if self.can232 and self.can232.isOpen():
            return 1
        return 0

    def getCanDeviceVersion(self):
        if self.can232:
            return self.can232.getDeviceVersion()
        return ""

    def runQueryCommand(self, cmd):
        if not self.isOpen():
            if not self.startModule():
                self.Log("Could not restart module after being failed.")
            else:
                self.Log("Module restarted after being failed.")
        
        id = 0
        type = 0
        for t in REGOCAN_DEVICES[self.REGOCAN_DEVICE][1]:
            if t[1] == cmd:
                id = t[0]
                type = t[2]
                break

        if type == 0:
            self.Log("ERROR: Invalid REGOCAN key: " + cmd)
            return ""

        res = self.runQueryId(id)
        if res != None:
            return convertRegoCanMessage(type, res)
        return ""

    def runQueryId(self, id):
        self.Debug("Running regoCanSerial query with id %d/0x%X" % (id, id))

        res = None
        if self.data_lock.lock_wait():
            if self.data.has_key(id):
                temp = self.data[id]
                if (time.time() - temp[0]) <= DATAVALID:
                    res = temp[1]
                else:
                    self.Debug("Data expired with id %d/0x%X" % (id, id))
            else:
                self.Debug("No data found with id %d/0x%X" % (id, id))
            self.data_lock.free()
        else:
            self.Log("Cannot get data lock.")
        return res   

        
###########################################################################

# Functions
    
def convertRegoCanMessage(type, data):
    if type == TYPE_UINT8:
        temp = ord(data[0])
        temp = "%d" % temp
        return temp
    elif type == TYPE_UINT16:
        temp = (ord(data[0]) << 8) + ord(data[1])
        temp = "%d" % temp
        return temp
    elif type == TYPE_TEMP16:
        temp = (ord(data[0]) << 8) + ord(data[1])
        # value 1023 is invalid temp
        if temp == 1023:
            return ""
        temp = convertRegoCanTemperature(temp)
        temp = "%.1f" % temp
        return temp

    return ""

# Constants for convertRegoCanTemperature function

# NTC-resistance-temperature-curve
NTC_CURVE_8016_CALCULATED = {}
TEMPERATURE_LOOKUP = {}
R0 = 4700.0
T0 = 25.0
TMIN = -50
TMAX = 150

A = -4.719299036435912E+03
B = -1.023642684798794E+03
C = -4.602444065258078E+01
D = -1.420793375371975E+03    

NTC_CURVE_8016 = { -50: 851300.0, -45: 269000.0, -40: 158100.0, -35: 111200.0, \
                   -30: 81700.0, -25: 60400.0, -20: 45100.0, -15: 33950.0, \
                   -10: 25800.0, -5: 19770.0, 0: 15280.0, 5: 11900.0, \
                   10: 9330.0, 15: 7370.0, 20: 5870.0, 25: 4700.0, \
                   30: 3790.0, 35: 3070.0, 40: 2510.0, 45: 2055.0, \
                   50: 1696.0, 55: 1405.0, 60: 1170.0, 65: 980.0, \
                   70: 824.0, 75: 696.0, 80: 590.0, 85: 503.0, \
                   90: 430.0, 95: 370.5, 100: 317.8, 105: 276.4, \
                   110: 240.7, 115: 205.5, 120: 180.7, 125: 161.0, \
                   130: 141.5, 135: 122.1, 140: 107.7, 145: 98.2, \
                   150: 83.9 }
       
def NtcCurve8016(Tn):
    t = int(math.floor(Tn / 5) * 5)
    if t > TMAX - 5:
        t = TMAX - 5
    if t < TMIN:
        t = TMIN
    c = NTC_CURVE_8016_CALCULATED[t]
    Rt = c[0] * R0
    return Rt * math.exp( (c[1]/100.0)*math.pow(t+273.15, 2.0)*(1.0/(Tn+273.15) - 1.0/(t+273.15)) )

def NtcCurve8016Reverse(Rn):
    lasti = -1
    lastr = -100000
    found = 0
    t = 0
    for i in range(TMIN,TMAX,5):
        r = NTC_CURVE_8016_CALCULATED[i][0] * R0
        if lastr < 0 and Rn > r:
            t = i
            found = 1
            break
        elif Rn > r:
            t = lasti
            found = 1
            break
        lasti = i
        lastr = r
    if not found:
        t = lasti
    c = NTC_CURVE_8016_CALCULATED[t]
    r = c[0] * R0
    t = t + 273.15
    
    try:
      tx = math.log(Rn/r) / ((c[1] / 100.0) * math.pow(t, 2.0)) + (1.0 / t)
      tx = 1.0 / tx - 273.15   
    except:
      tx = float(TMAX)
    if tx > TMAX:
      tx = float(TMAX)
    if tx < TMIN:
      tx = float(TMIN)
    return tx

def valueToResistance(val):
    return A * val / (B + val) - C * val / (D + val)

if len(TEMPERATURE_LOOKUP.keys()) <= 0:
    for i in range(TMIN,TMAX,5):
        tx = i + 5
        NTC_CURVE_8016_CALCULATED[i] = [NTC_CURVE_8016[i] / R0, 100.0 * math.log(NTC_CURVE_8016[tx]/NTC_CURVE_8016[i]) / (math.pow(i + 273.15, 2.0) * (1.0/(tx + 273.15) - 1.0/(i + 273.15)))]
    for i in range(0, 1023):
        r = valueToResistance(i)
        TEMPERATURE_LOOKUP[i] = round(NtcCurve8016Reverse(r), 1)

def convertRegoCanTemperature(value):
    try:
        return TEMPERATURE_LOOKUP[value]
    except:
        return float(TMIN)

