#!env python
# -*- coding: iso-8859-1 -*-
###########################################################################
# 
# File:            raspberryPiGPIO.py
#
# License:         Donationware, see attached LICENSE file for more 
#                  information
#
# Author:          Olli Lammi (olammi@iki.fi)
#
# Version:         1.6d
#
# Date:            28.12.2014
#
# Description:     Library to communicate with taloLogger raspberryGpioThread.
#                  
# Requirements:    Python interpreter 2.6 or newer (www.python.org)
#
# Version history: ** 28.12.2014 v1.6d (Olli Lammi) **
#                  First version.
#
###########################################################################

# Imports

import sys, os 
import string
import time
import subprocess

from modules.core import log
from modules.core import threads

###########################################################################

# Classes

PHASE_TERMINATED = 0
PHASE_STARTED = 1
PHASE_INITIALIZED = 2
PHASE_TERMINATING = 3

class GPIOSource(threads.Thread, log.Logging):
    def __init__(self):
        threads.Thread.__init__(self)
        log.Logging.__init__(self, 'GPIOSource')
        self.DATAPOINTS = {}
        
        self.errormsg = ""
        self.query = []
        self.query_lock = threads.Lock()
        self.data = {}
        self.data_ready = 0
        self.data_lock = threads.Lock()
        self.phase = PHASE_TERMINATED
        
    def initModule(self):
        if self.phase != PHASE_TERMINATED or not self.isTerminated():
            self.Log("Cannot initialize non terminated GPIOSource thread.")
            return 0
        
        if not self.start():
            self.Log("Cannot start GPIOSource thread.")
            return 0
        self.phase = PHASE_STARTED
        
        while self.phase == PHASE_STARTED:
            time.sleep(0.1)
        
        if self.phase == PHASE_INITIALIZED:        
            self.Log("Initialized GPIO thread.")
            return 1

        self.Log("Error initializing GPIO source: " + self.errormsg)
        return 0

    def closeModule(self):
        try:
            self.Debug("Closing GPIO thread.")
            if not self.hasTerminated():
                self.Debug("Terminating GPIO thread.")
                self.terminate()
                self.wait_sleeping_timeout(5.0)
        except:
            self.Debug("Error closing GPIO thread.")

    def setDatapoints(self, pointmap):
        self.DATAPOINTS = pointmap

    def run(self):
        sb = None
        command = findRaspberryGpioLauncher()
        
        try:              
            while self.isRunning():
                if sb == None:
                    try:      
                        sb = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                    except:
                        self.Log("ERROR: Cannot run raspberryGpioThread.py.")
                        self.errormsg = "Cannot run raspberryGpioThread.py."
                        self.setFail()
                        self.phase = PHASE_TERMINATING
                        break

                if self.isRunning() and self.phase == PHASE_STARTED:
                    # initialize
                    for dpkey in self.DATAPOINTS.keys():
                        temp = self.communicateWithSubprocess(sb, "input %d %s" % (self.DATAPOINTS[dpkey][0], self.DATAPOINTS[dpkey][1]))
                        if temp != 'OK':
                            self.errormsg = "Error configuring input: %s" % dpkey
                            self.setFail()
                            break
                    temp = self.communicateWithSubprocess(sb, "init")
                    if temp != 'OK':
                        self.errormsg = "Error initializing GPIO: " + temp
                        self.setFail()
                        break
                    self.phase = PHASE_INITIALIZED
                               
                if self.isRunning() and self.phase == PHASE_INITIALIZED:
                    temp = None
                    if self.query_lock.lock_wait():
                        if len(self.query) > 0:
                            temp = self.query[0]
                            self.query = self.query[1:]
                        self.query_lock.free()
                    if self.isRunning() and temp != None:
                        while self.data_lock.lock_wait():
                            if self.data_ready == 0:
                                for key in temp:
                                    temps = self.communicateWithSubprocess(sb, "read %d" % self.DATAPOINTS[key][0])
                                    if len(temps) >= 5 and temps[:5] == 'ERROR':
                                        self.Log("Received error for GPIO input %s: %s" % (key, temps))
                                        self.data[key] = ""
                                    else:
                                        self.data[key] = temps
                                self.data_ready = 1
                                self.data_lock.free()
                                break
                            self.data_lock.free()
                            time.sleep(0.1)
                                     
                if self.isRunning():
                    time.sleep(0.1)

        except Exception, e:
            self.Log("Exception: " + e.__str__())
            self.errormsg = "Exception"
            self.phase = PHASE_TERMINATING
            self.setFail()
        except IOError, ioe:
            self.Log("IOError: " + ioe.__str__())
            self.errormsg = "Exception"
            self.phase = PHASE_TERMINATING
            self.setFail()

        if sb != None:
            if self.phase != PHASE_TERMINATING:
                self.Debug("Sending exit command.")
                sb.stdin.write('exit' + chr(10))
                sb.stdin.flush()
            else:
                self.Debug("Terminating GPIO launcher.")
                try:
                    sb.terminate()
                except:
                    pass
            sb = None

        self.phase = PHASE_TERMINATED       
        self.Log("GPIOSource thread stopped.")


    def runQueryCommand(self, cmds):
        dataresult = {}
        for key in cmds:
            dataresult[key] = ""

        if self.hasTerminated() and self.isFailed():
            if not self.initModule():
                self.Log("Could not restart module after being failed.")
                return dataresult
            else:
                self.Log("Module restarted after being failed.")

        if not self.phase == PHASE_INITIALIZED:
            self.Log("Error, thread not initialized.")
            return dataresult

        if self.query_lock.lock_wait():
            self.query.append(cmds)
            self.query_lock.free()
        else:
            self.Log("Error, cannot add query.")
            return dataresult
                                
        # wait for data complete
        while self.data_lock.lock_wait():
            if self.data_ready:
                for key in cmds:
                    dataresult[key] = self.data[key]
                self.data_ready = 0
                self.data = {}
                self.data_lock.free()
                break
            self.data_lock.free()

        return dataresult


    def communicateWithSubprocess(self, sb, cmd):
        self.Debug("Sending command: " + cmd)
        try:
            sb.stdin.write(cmd + chr(10))
            sb.stdin.flush()
        except:
            self.Log("Failed sending command.")
            self.phase = PHASE_TERMINATING
            self.setFail()
            return ""
        
        buff = ""
        while 1:
            tempc = sb.stdout.read(1)
            if not tempc:
                break
            elif tempc == chr(10):
                self.Debug("Received: " + buff)
                return buff
            else:
                buff = buff + tempc
        return buff

###########################################################################

# Functions

def findRaspberryGpioLauncher():
    # FIXME!
    return "/home/talo/bin/raspberry_gpio_launcher"
