#!env python
# -*- coding: iso-8859-1 -*-
###########################################################################
# 
# File:            telldusCore.py
#
# License:         Donationware, see attached LICENSE file for more 
#                  information
#
# Author:          Olli Lammi (olammi@iki.fi)
#
# Version:         1.6a
#
# Date:            12.05.2014
#
# Description:     Library to communicate with wireless sensors and 
#                  Telldus Tellstick Duo using telldus-core-library.
#                  
# Requirements:    Python interpreter 2.5 or newer (www.python.org)
#
# Version history: ** 12.05.2014 v1.6a (Olli Lammi) **
#                  First version.
#
###########################################################################

# Imports

import sys, os 
import string, ctypes

from modules.core import log
from modules.utils import utils


###########################################################################

# Constants

SENSOR_DATATYPES = { 'TEMPERATURE': 1, 'HUMIDITY': 2, 'RAINRATE': 4, \
                     'RAINTOTAL': 8, 'WINDDIRECTION': 16, 'WINDAVERAGE': 32, \
                     'WINDGUST': 64 }

DATA_LEN = 256

###########################################################################

# Classes

class TelldusSource(log.Logging):
    def __init__(self):
        log.Logging.__init__(self, 'TelldusSource')
        self.DATAPOINTS = {}
        self.libtd = None
        
    def initLibrary(self):
        self.libtd = ctypes.cdll.LoadLibrary('libtelldus-core.so.2')
        if self.libtd == None:
            return 0
        
        try:
            self.Debug("Initializing telldus-core library.")
            self.libtd.tdInit()
        except:
            self.Log("Cannot initialize telldus-core library.")
            self.libtd = None
            return 0

        self.Log("Initialized telldus-core library.")
        return 1

    def closeLibrary(self):
        try:
            self.Debug("Closing telldus-core library.")
            self.libtd.tdClose()
        except:
            self.Debug("Error closing libtelldus-core.")

        self.libtd = None

    def setDatapoints(self, pointmap):
        self.DATAPOINTS = pointmap

    def runQueryCommand(self, cmds):
        if self.libtd == None:
            self.Log("ERROR: No libtelldus-core.so.2 library available.")
            return {}
        
        sensordata = []
        protocol = ctypes.create_string_buffer(DATA_LEN)
        model = ctypes.create_string_buffer(DATA_LEN)
        sensorId = ctypes.c_int(0)
        dataTypes = ctypes.c_int(0)
        value = ctypes.create_string_buffer(DATA_LEN)
        tstamp = ctypes.c_int(0)
        
        try:
            while self.libtd.tdSensor(protocol, DATA_LEN, model, DATA_LEN, \
                                      ctypes.byref(sensorId), ctypes.byref(dataTypes)) == 0:
                self.Debug("Found sensor %s:%s:%d" % (protocol.value, model.value, sensorId.value))
    
                for dtype in SENSOR_DATATYPES.keys():
                    dtid = SENSOR_DATATYPES[dtype]
                    if dataTypes.value & dtid:
                        self.Debug("  Reading sensor data type: %s" % dtype)
                        try:
                            self.libtd.tdSensorValue(protocol, model, sensorId, dtid, \
                                                     value, DATA_LEN, ctypes.byref(tstamp))
                            sensordata.append([protocol.value, model.value, sensorId.value, dtid, value.value])
                        except:
                            self.Log("Error reading Telldus sensor value.")
        except:
            self.Debug("Error reading Telldus sensors.")
                
        self.Debug("Got %d sensor values from telldus-core." % len(sensordata))

        if len(sensordata) <= 0:
            self.Log("No sensor data from telldus-core.")
            return {}
        
        dataresult = {}
                
        for cmd in cmds:
            if not self.DATAPOINTS.has_key(cmd):
                self.Log("ERROR: Undefined data point: " + cmd)
                continue
            item = self.DATAPOINTS[cmd]
                                
            protocol = item[0]
            model = item[1]
            id = item[2]
            datatype = item[3]                    

            if datatype not in SENSOR_DATATYPES.keys():                
                self.Log("ERROR: Invalid or missing data type for data point: " + cmd)
                continue
            
            found = 0
            for sensoritem in sensordata:
                if sensoritem[0] == protocol and sensoritem[1] == model and \
                    sensoritem[2] == id and SENSOR_DATATYPES[datatype] == sensoritem[3]:
                    found = 1
                    dataresult[cmd] = sensoritem[4]

            if not found:            
                self.Log("ERROR: Sensor data not found: " + cmd)
                continue
                
        return dataresult
        

###########################################################################

# Functions
