#!env python
# -*- coding: iso-8859-1 -*-
###########################################################################
# 
# File:            storeDbSQLite.py
#
# License:         Donationware, see attached LICENSE file for more 
#                  information
#
# Author:          Olli Lammi (olammi@iki.fi)
#
# Version:         1.0e
#
# Date:            15.01.2013
#
# Description:     Data Store class to store logged data values into a
#                  SQLite file based database.
#                  
# Requirements:    Python interpreter 2.5 or newer (www.python.org)
#                  (tested with 2.7)
# 
#                  MySQL-python library (MySQLdb) 
#                  (http://sourceforge.net/projects/mysql-python)
#
# Version history: ** 22.10.2010 v1.0a (Olli Lammi) **
#                  SQLite support added. 
#
#                  ** 10.02.2011 v1.0b (Olli Lammi) **
#                  Interface changes. 
#
#                  ** 26.10.2011 v1.0c (Olli Lammi) **
#                  Added store filter init. 
#
#                  ** 07.01.2013 v1.0d (Olli Lammi) **
#                  Support for log time passing. 
#
#                  ** 15.01.2013 v1.0e (Olli Lammi) **
#                  Added support for native db schema v1. 
#
###########################################################################

# Imports

import sys, os 
import string

from modules.core import store
from modules.core import configuration

# sqlite3 module, not loaded until init of DBStoreSQLite class instance
SQLiteDb = None


###########################################################################

# Constants

CREATE_TALO_DATA_1 = "CREATE TABLE IF NOT EXISTS talo_data (" + \
                     "id INTEGER PRIMARY KEY AUTOINCREMENT, " + \
                     "time INTEGER NOT NULL, " + \
                     "position_id INTEGER NOT NULL, " + \
                     "value NUMERIC)"

CREATE_TALO_POSITIONS_1 = "CREATE TABLE IF NOT EXISTS talo_positions (" + \
                     "id INTEGER PRIMARY KEY AUTOINCREMENT, " + \
                     "position_name TEXT NOT NULL UNIQUE)"

CREATE_TALO_DATA_TIME_INDEX_1 = "CREATE INDEX IF NOT EXISTS i_talo_data_time ON talo_data (time)"


###########################################################################

# Classes

class DBStoreSQLite(store.Store):
    def __init__(self, modname):
        store.Store.__init__(self, modname)

        global SQLiteDb
        SQLiteDb = __import__('sqlite3') 

        self.POSITIONS = {}

    def handleConfiguration(self, conf):
        if not self.initStoreFilters(conf):
            return (0, "Invalid store filter.")
        
        self.DB_NAME = conf.getValue('NAME', '', self.getModuleName())
        self.DB_TABLE = conf.getValue('TABLE', '', self.getModuleName())
        self.DB_TIMECOL = conf.getValue('TIMECOL', '', self.getModuleName())        
                
        self.DB_SCHEMA = conf.getValue('DB_SCHEMA_VERSION', '', self.getModuleName())
        if len(self.DB_SCHEMA) <= 0 or int(self.DB_SCHEMA) <= 0:
            self.DB_SCHEMA = 0
        elif len(self.DB_SCHEMA) > 0:
            self.DB_SCHEMA = int(self.DB_SCHEMA)
        if self.DB_SCHEMA == 1:
            self.DB_TABLE = 'talo_data'
            self.DB_TIMECOL = 'time'
        elif self.DB_SCHEMA != 0:
            return (-1, 'Invalid DB schema version.')

        return (1, '')

    def initConfiguration(self):
        status = 0
        sqlstmt = ""
        try:
            db = SQLiteDb.connect(self.DB_NAME)
            cur = db.cursor()
            if self.DB_SCHEMA == 1:
                sqlstmt = CREATE_TALO_DATA_1
                cur.execute(sqlstmt)
                sqlstmt = CREATE_TALO_POSITIONS_1
                cur.execute(sqlstmt)
                sqlstmt = CREATE_TALO_DATA_TIME_INDEX_1
                cur.execute(sqlstmt)
            sqlstmt = "SELECT " + self.DB_TIMECOL + " FROM " + self.DB_TABLE + " LIMIT 1"
            cur.execute(sqlstmt)
            cur.close()
            db.close()
            status = 1
        except:
            self.Log("ERROR: Error in database operation, SQL: " + sqlstmt)
            status = 0
        if status:
            self.initPositions()
            return (1, '')
        else:
            return (0, 'DBStoreSQLite: Errors with database access.')

 
    def initPositions(self):
        if self.DB_SCHEMA == 1:
            self.POSITIONS = {}
            sqlstmt = ""
            try:
                db = SQLiteDb.connect(self.DB_NAME)
                cur = db.cursor()
                sqlstmt = "SELECT id, position_name FROM talo_positions"
                cur.execute(sqlstmt)
                res = cur.fetchall()
                cur.close()
                db.close()
                for row in res:
                    self.POSITIONS[row[1]] = row[0]
            except:
                self.Log("ERROR: Error in database operation, SQL: " + sqlstmt)
 
    
    def insertDataImpl(self, timeval, values):
        if self.DB_SCHEMA == 1:
            if len(self.POSITIONS.keys()) <= 0:
                self.initConfiguration()
                
            status = 1
            for val in values:
                if not self.POSITIONS.has_key(val[0]):
                    status = 0
                    break
            if status == 0:
                self.initPositions()
                toadd = []
                for val in values:
                    if not self.POSITIONS.has_key(val[0]):
                        toadd.append(val[0])
                
                try: 
                    db = SQLiteDb.connect(self.DB_NAME)
                    cur = db.cursor()
                    for col in toadd:
                        sqlstmt = "INSERT INTO talo_positions (position_name) VALUES ('"
                        sqlstmt = sqlstmt + col
                        sqlstmt = sqlstmt + "')"
                        cur.execute(sqlstmt)
                    cur.close()
                    db.commit()
                    db.close()
                except:
                    self.Log("ERROR: Error in database operation, SQL: " + sqlstmt)
                    self.POSITIONS = {}
                    return 0
                self.initPositions()
        
        sqlstmt = ""
        try:
            # time as integer as seconds from epoch
            timestr = "DATETIME(%d, 'unixepoch')" % (timeval, )

            db = SQLiteDb.connect(self.DB_NAME)
            if self.DB_SCHEMA == 1:
                for val in values:
                    cur = db.cursor()
    
                    posid = self.POSITIONS[val[0]]
                            
                    sqlstmt = "INSERT INTO " + self.DB_TABLE + " "
                    sqlstmt = sqlstmt + "(" + self.DB_TIMECOL + ", position_id, value) VALUES ("
                    sqlstmt = sqlstmt + "%s, %d, " % (timestr, posid)
                    sqlstmt = sqlstmt + checkDBValue(val[1]) + ")"
        
                    self.Debug("Inserting data to database using SQL: " + sqlstmt)
                    cur.execute(sqlstmt)
                    cur.close()
            else:                            
                cur = db.cursor()
    
                columnstr = ""
                valuestr = ""
                for item in values:
                    if len(columnstr) > 0:
                        columnstr = columnstr + ", "
                    if len(valuestr) > 0:
                        valuestr = valuestr + ", "
                    columnstr = columnstr + item[0]
                    valuestr = valuestr + checkDBValue(item[1])
                    
                sqlstmt = "INSERT INTO " + self.DB_TABLE + " "
                sqlstmt = sqlstmt + "(" + self.DB_TIMECOL + ", "
                sqlstmt = sqlstmt + columnstr + ") VALUES (" + timestr + ", "
                sqlstmt = sqlstmt + valuestr + ")"
    
                self.Debug("Inserting data to database using SQL: " + sqlstmt)
                cur.execute(sqlstmt)
                cur.close()
            db.commit()
            db.close()
        except:
            self.Log("ERROR: Error in database operation, SQL: " + sqlstmt)
            self.POSITIONS = {}
            return 0
        return 1

    @staticmethod
    def getModuleTypeName():
        return 'SQLITEDB'

    @staticmethod
    def getAllowedConfigurationKeys():
        return (['NAME', 'TABLE', 'TIMECOL', 'DB_SCHEMA_VERSION'], [])


###########################################################################

# Functions

def checkDBValue(val):
    if (val and len(val) > 0):
        return val
    return "NULL"

