#!env python
# -*- coding: iso-8859-1 -*-
###########################################################################
# 
# File:            taloLogger.py
#
#
# License:         Donationware, see attached LICENSE file for more 
#                  information
#
#
HELP_TEXT = """
# Usage:           python taloLogger.py --help
#                  python taloLogger.py [-d] [--nodaemon] [-v] [-l] [-f configuration_file]
#
#                  Options (options override the configuration file):
#                    --help     - Prints out a help and exits.
#
#                    -d         - Runs the application in daemon mode. Command will
#                                 return to shell and a daemon will be forked to
#                                 run in the background. 
#
#                    --nodaemon  - No daemon mode. Even if -d option or configuration
#                                 file directives would make taloLogger run in daemon
#                                 mode, this overrides them all and prevents forking.
#
#                    -v         - Enable debug/verbose logging.
#
#                    -l         - Console log mode. 
#
#                    -f configuration_file - Use the given filename as configuration
#                                 file for the application.
#
#                  Parameters:
#                                 None.
#
#                  Configuration:
#                                 See file taloLogger.conf for example configuration
#                                 and configuration keys.
#
"""
#
# Author:          Olli Lammi (olammi@iki.fi)
#
# Version:         1.6f
#
# Date:            XX.01.2015
#
# Functions:       -
#                   
# Description:     House automation logger application for Python. The primary
#                  driver has been to produce a logger application to read Ouman
#                  control device and 1-wire sensor values and log them 
#                  periodically to MySql database. The application has been
#                  extended with various data source and data storage 
#                  modules.
#                  
#                  Can be extended to log to different data storages: file,
#                  rrdtool, database, etc. Suggest a new data storage for the 
#                  author if you need a specific module.
#
#                  Currently supported data sources:
#                    + Ouman 200 and 686 series controllers using serial connection
#                    + Ouman EH800 controllers using ethernet connection
#                    + Thermia heat pump controllers using serial connection
#                    + Rego600 series controllers using serial connection
#                    + Rego800 and 1000 controllers with Can232/CanUSB interface module.
#                    + Ekowell heat pump controllers using serial connection
#                    + Nibe heat pump controllers using serial connection
#                    + ThermIQ interface (www.thermiq.net) connected to Thermia 
#                      or Danfoss heat pumps
#                    + Siemens SmartWeb module using network (web) connection
#                    + OneWire bus appliances and sensors using either OWFS,
#                      DigiTemp or EDS OW-SERVER-ENET-2 interface 
#                    + Any runnable shell command or program
#                    + Telldus Tellstick Duo wireless sensor devices, 
#                      using telldus-core-library
#                    + Raspberry Pi GPIO inputs (only on Raspberry Pi with 
#                      taloLoggerPi)
#
#                  Currently supported data storages:  
#                    + text file
#                    + RRD database file
#                    + MySQL database tables
#                    + SQLite3 database files
#
# Requirements:    Python interpreter 2.6 or newer (www.python.org).
#                  Series 3.X Python not supported.
# 
#                  For Ouman, Thermia, Rego600, RegoCan, Ekowell, Nibe  
#                  and ThermIQ controller logging:
#                    Python Serial Port Extension: pySerial  
#                    (http://pyserial.wiki.sourceforge.net/pySerial)
#
#                  For MySQL database data storage:
#                    MySQL-python library (MySQLdb) 
#                    (http://sourceforge.net/projects/mysql-python)
#
#                  For 1-wire data sources one of below:
#                    OWFS (http://www.owfs.org/) 
#                    DigiTemp (http://www.digitemp.com/)
#
#                  For RRDtool data storage:
#                    RRDtool (http://oss.oetiker.ch/rrdtool/)
#
#                  For Telldus Tellstick Duo wireless sensor sources:
#                    telldus-core (http://developer.telldus.se/doxygen/)
#
#                  For Raspberry Pi GPIO sources:
#                    taloLoggerPi (http://olammi.iki.fi/sw/taloLoggerPi/)
#
# Version history: ** 24.02.2009 v0.9a (Olli Lammi) **
#                  First beta version for testers. 
#
#                  ** 24.02.2009 v0.9b (Olli Lammi) **
#                  Added support for rrdtool data store. 
#
#                  ** 25.02.2009 v0.9c (Olli Lammi) **
#                  Changed logging. Changed rrdtool support to be
#                  not dependent on external rrdtool-python library.
#                  Changed module imports not to import requirement
#                  modules if a module is not enabled at the configuration. 
#
#                  ** 02.03.2009 v0.9d (Olli Lammi) **
#                  Added verbose logging and daemon configuration directive.
#                  Added pattern support for file data store. Changed the
#                  way oumanSerial accesses the serial port and added 
#                  file locking for the port access.
#
#                  ** 03.03.2009 v0.9e (Olli Lammi) **
#                  Changed the way how empty measured values are handled.
#
#                  ** 06.04.2009 v1.0a (Olli Lammi) **
#                  First release candidate version. Moved configuration objects to the
#                  modules. If no values are available, completely empty data 
#                  is not stored to datastores. Added preliminary config checks.
#                  Added support for DigiTemp. 
#
#                  ** 06.04.2009 v1.0b (Olli Lammi) **
#                  Correction to DigiTemp tool output handling.
#
#                  ** 02.06.2009 v1.0c (Olli Lammi) **
#                  Added Ouman device profile for 202G controller (Geopro GS-pumps)
#
#                  ** 29.04.2010 v1.1a (Olli Lammi) **
#                  Added Thermia serial module.
#
#                  ** 04.09.2010 v1.1b (Olli Lammi) **
#                  Corrections to Thermia functionality (bugs found in beta testing)
#
#                  ** 22.10.2010 v1.1c (Olli Lammi) **
#                  Added SQLite support.
#
#                  ** 17.08.2010 v1.2a (Olli Lammi) **
#                  Added scripts to get and set Ouman device parameters.
#                  Support for EH202S parameters.
#
#                  ** 13.09.2010 v1.2b (Olli Lammi) **
#                  Initial support for parameter changes for Ouman EH203GT.
#
#                  ** 15.11.2010 v1.2c (Olli Lammi) **
#                  Extended the support for Ouman EH203GT parameters.
#
#                  ** 16.11.2010 v1.3a (Olli Lammi) **
#                  Added Rego serial module.
#
#                  ** 18.11.2010 v1.3b (Olli Lammi) **
#                  Rego module bug corrections.
#
#                  ** 08.02.2011 v1.3c (Olli Lammi) **
#                  Added Ekowell serial module.
#
#                  ** 08.02.2011 v1.4a (Olli Lammi) **
#                  Support for multiple data store and source modules in 
#                  the configuration. Interface changes.
#
#                  ** 11.04.2011 v1.4c (Olli Lammi) **
#                  Added Nibe serial module. Added support for SmartWeb.
#                  Small correction to OWFS "nan" handling. 
#
#                  ** 18.10.2011 v1.4e (Olli Lammi) **
#                  Corrected error handling in smartWeb module.
#                  Restructured modules. Corrections to Nibe serial module.
#                  Added support for ThermIQ serial module. 
#
#                  ** 26.10.2011 v1.4f (Olli Lammi) **
#                  Data store filtering configuration.
#
#                  ** 01.11.2011 v1.4g (Olli Lammi) **
#                  Fault tolerance to nibeSerial.
#
#                  ** 07.11.2011 v1.4h (Olli Lammi) **
#                  Use script or executable as data source.                   
#
#                  ** 13.02.2012 v1.4i (Olli Lammi) **
#                  Redesigned the source interfaces and logic.                   
#
#                  ** 15.01.2013 v1.5a (Olli Lammi) **
#                  Values now logged with timestamp of time when queuing
#                  the event. Added regoCanSerial module support.
#                  Added dummy source support.                   
#
#                  ** 31.01.2013 v1.5b (Olli Lammi) **
#                  Added support for EDS OW-SERVER-ENET-2.
#                  Added support for Ouman EH800.
#                  Requirement updated to Python 2.6 or newer.
#
#                  ** 24.03.2013 v1.5c (Olli Lammi) **
#                  Enhanced MEASURE syntax verification.
#                  Added support for virtual measures.
#                  For other changes, see CHANGELOG.
#
#                  ** 12.05.2014 v1.6a (Olli Lammi) **
#                  Increased ThermIQ serial module timeout.
#                  Added support for Telldus Tellstick Duo wireless 
#                  sensor devices.
#
#                  ** 14.05.2014 v1.6b (Olli Lammi) **
#                  Fixed data saving issues when timeout or invalid data. 
#                  Invalid data will not stop logging. Discarded data 
#                  will be written to log.
#
#                  ** 01.12.2014 v1.6b (Olli Lammi) **
#                  Added ON/OFF-type datapoint handling to Ouman EH800. 
#
#                  ** 02.12.2014 v1.6b (Olli Lammi) **
#                  Works now again in Windows environment. 
#
#                  ** 02.12.2014 v1.6b (Olli Lammi) **
#                  Changed shellSource-module to do regexp searching vs. matching.
#                  This improves searching in multiline command outputs. 
#                  Also enabled DOTALL flag (regexp . matches also newlines). 
#                  Added data points (GT1 and GT3 target/on/off values) to 
#                  regoSerial module.
#
#                  ** 10.12.2014 v1.6c (Olli Lammi) **
#                  Released 1.6b beta as v1.6c.
#
#                  ** 11.12.2014 v1.6d (Olli Lammi) **
#                  Module init writes to log.
#
#                  ** 28.12.2014 v1.6d (Olli Lammi) **
#                  Support for Raspberry Pi GPIO inputs.
#                  Added --nodaemon option to prevent accidental forking
#                  configuration in taloLoggerPi.
#
#                  ** 14.01.2015 v1.6e (Olli Lammi) **
#                  Changed directory structure and module configuration.
#                  Added modbus sources and Nibe Modbus. Test version.
#                  Added Nibe Bus RS485 Serial and UDP module.
#
###########################################################################

# Imports

import sys, os 
import string, re, time  
import traceback
import platform, urllib2

from modules.core import threads 
from modules.core import configuration
from modules.core import dataSource
from modules.core import log
from modules.core import store
from modules.core import virtualMeasures

from modules.core import sourceDummy

from modules.datasources import *

from modules.datastores import *

###########################################################################

# Constants

VERSION = "v1.6f"

DEFAULT_LOG_INTERVAL = 120
DEFAULT_LOG_INTERVAL_LIMIT = 5


###########################################################################

# Globals

main_terminated = 0

LOG = None


###########################################################################

# Classes

class TaloLoggerThread(threads.Thread, log.Logging, dataSource.DataSourceListener):
    def __init__(self, meas, virt, sources, stores):
        threads.Thread.__init__(self)
        log.Logging.__init__(self, 'TaloLoggerThread')

        self.inqueue = []
        self.inqueue_item = []
        self.inq_lock = threads.Lock()
        self.datasourcelist = sources
        self.storelist = stores
        self.measures = meas
        self.virtuals = virt
        self.terminated = 0
        self.prevresult = {}

    def terminate(self):
        self.terminated = 1

    def isterminated(self):
        return self.terminated

    def dataReceived(self, moduleid, data, queuets):
        if self.inq_lock.lock_wait():
            haslock = True
            if len(self.inqueue_item) > 0:
                for item in self.inqueue_item:
                    if item[2] == queuets:                         
                        for key in data.keys():
                            tkey = moduleid + '.' + key
                            if tkey in item[1]:
                                item[3][tkey] = data[key]
                            else:
                                self.Log("Received not queried data: " + tkey)

                status = 1
                while len(self.inqueue_item) > 0 and status == 1: 
                    for k in self.inqueue_item[0][1]:
                        if k not in self.inqueue_item[0][3]:
                            status = 0
                    if status:
                        item = self.inqueue_item[0]
                        self.inqueue_item = self.inqueue_item[1:]
                        haslock = False
                        self.inq_lock.free()
    
                        tempresult = item[3]
                        for k in tempresult.keys():
                            if len(tempresult[k]) <= 0:
                                self.Log("Received none or unknown data for command " + k + ".")                    
                        self.handleResult(item)
                        
                        if self.inq_lock.lock_wait():
                            haslock = True
                        else:
                            break
            else:
                self.Log("Received not queried data.")
            if haslock:
                self.inq_lock.free()
        
    def handleResult(self, item):
        if item[0] == 'log':
            self.handleLogResult(item[2], item[3])
        else:
            data = item[3]
            self.Log("Received unknown item type " + item[0]);
            for k in data.keys():
                self.Log("  " + k + ":" + data[k])
        
    def handleLogResult(self, timeval, data):
        self.Debug("Storing logged data with " + repr(len(data)) + " points.")
        # map the result data to keys and values
        result = []
        resultdict = {}
        for item in self.measures:
            temps = data[item[1]]
            resultdict[item[0]] = temps
            result.append([item[0], temps])
            
        virtresult = virtualMeasures.HandleVirtuals(self, self.virtuals, timeval, resultdict, self.prevresult)
        for k in virtresult.keys():
            resultdict[k] = virtresult[k]
            result.append([k, virtresult[k]])
        self.prevresult = resultdict
        self.prevresult['%TIME%'] = timeval    
        
        num = 0
        for item in result:
            if len(item[1]) > 0:
                num = num + 1
        
        if num > 0:
            if len(self.storelist) > 0:
                for store in self.storelist:
                    store.insertData(timeval, result)
            else:
                temps = ""
                for item in result:
                    temps = temps + "\n    " + item[0] + ": " + item[1]
                self.Log(temps)
        else:
            self.Log("Completely empty result, not storing data.")

    def queue(self, type, val, timeval):
        if self.inq_lock.lock_wait():
            self.inqueue_item.append([type, val, timeval, {}])
            temp = {}
            for v in val:
                temp2 = string.split(v, '.', 1)
                if not temp.has_key(temp2[0]):
                    temp[temp2[0]] = [ temp2[1] ]
                else:
                    temp[temp2[0]].append(temp2[1])
            for k in temp.keys():
                self.inqueue.append([k, temp[k], timeval])
            self.inq_lock.free()

    def run(self):
        while not self.terminated:
            try:
                # check if there is something to do
                if self.inq_lock.lock_wait():
                    temp = None
                    if len(self.inqueue) > 0:
                        temp = self.inqueue[0]
                        self.inqueue = self.inqueue[1:]
                    self.inq_lock.free()                    
                    
                    if temp != None:
                        fnd = False                    
                        isok = False
                        for s in self.datasourcelist:
                            if s.getServiceName() == temp[0]:
                                fnd = True
                                if s.runDataSourceQueryCommandAsync(self, temp[0], temp[1], temp[2]):
                                    isok = True
                                else:
                                    self.Log("Unable to start data source query for module " + temp[0] + ".")
                        if not fnd:
                            self.Log("Unknown command source " + temp[0])
                        if not isok:
                            tempdata = {}
                            for key in temp[1]:
                                tempdata[key] = ""
                            self.dataReceived(temp[0], tempdata, temp[2])
                    else:
                        time.sleep(1)
            except Exception, e:
                traceback.print_exc(file=sys.stdout)
                self.Log("Exception: " + e.__str__())
                self.terminated = 1


class ReleasedVersionChecker(threads.Thread, log.Logging):
    def __init__(self):
        threads.Thread.__init__(self)
        log.Logging.__init__(self, 'ReleasedVersionChecker')
        
    def run(self):
        url = 'http://olammi.iki.fi/sw/taloLogger/current_release.txt'
        res = ''
        try:
            USERAGENT = "taloLogger/" + VERSION + " (Python/" + platform.python_version() + "; Sys/" + platform.platform() + ")"
            req = urllib2.Request(url)
            req.add_header("User-Agent", USERAGENT)
            f = urllib2.urlopen(req)
            res = f.readlines()
            f.close()
        except:
            return
    
        if len(res) <= 0:
            return
    
        res = string.strip(string.split(res[0], '\n')[0])
        if len(res) <= 0:
            return
    
        if VERSION < res:
            self.Log("\n######################################################################\n" + \
                     "#    There is a newer release version of taloLogger available in \n" + \
                     "# \n" + \
                     "#                  http://olammi.iki.fi/sw/taloLogger/\n" + \
                     "# \n" + \
                     "#    Current version: %s    Available release version: %s\n" % (VERSION, res) + \
                     "######################################################################")



###########################################################################

# Functions

def CheckPythonVersion():
    version = platform.python_version_tuple()
    # Require Python 2.6 or newer.
    try:
        if int(version[0]) != 2 or int(version[1]) < 6:
            print "ERROR: Invalid Python version to run taloLogger. Required Python series 2 newer or 2.6."
            print "       Detected Python version: %s.%s.%s" % version
            sys.exit(1)
    except:
        print "Error determining Python version."
    return

def HUPhandler(signum, frame):
  global main_terminated, LOG
  LOG.log("Terminating due to SIGHUP.")
  main_terminated = 1
  return

def TERMhandler(signum, frame):
  global main_terminated, LOG
  LOG.log("Terminating due to SIGTERM.")
  main_terminated = 1
  return

def INThandler(signum, frame):
  global main_terminated, LOG
  LOG.log("Terminating due to SIGINT.")
  main_terminated = 1
  return

def GetModuleTypeAndName(str):
    if string.find(str, ':') < 0:
        return ('', '')
    [mtype, mname] = string.split(str, ':', 1)
    mtype = string.strip(mtype)
    mname = string.strip(mname)
    return (mtype, mname)

def GetDataSourceClass(mtype):
    for cls in dataSource.DataSource.__subclasses__():
        if issubclass(cls, configuration.Configurable):
            if mtype == cls.getModuleTypeName():
                return cls
    return None

def GetDataStoreClass(mtype):
    for cls in store.Store.__subclasses__():
        if issubclass(cls, configuration.Configurable):
            if mtype == cls.getModuleTypeName():
                return cls
    return None


def CheckReleasedVersion():
    checker = ReleasedVersionChecker()
    checker.start()
      
              
####### main ##########################################################

def main():
  global main_terminated, LOG

  CheckPythonVersion()

  CONFIGURATION_FILE = "taloLogger.conf"

  nodaemon = 0
  daemon = 0
  logconsole = 0
  verbose = 0

  # handle command line parameters
  i = 1
  while i < len(sys.argv):
      if sys.argv[i] == '--help':
          Help()
          sys.exit(1)
      elif sys.argv[i] == '-l':
          logconsole = 1
      elif sys.argv[i] == '-v':
          verbose = 1
      elif sys.argv[i] == '-d':
          daemon = 1
      elif sys.argv[i] == '--nodaemon':
          nodaemon = 1
      elif  sys.argv[i] == '-f' and i+1 < len(sys.argv):
          i = i + 1
          CONFIGURATION_FILE = sys.argv[i]
      else:
          Usage()
          sys.exit(1)

      i = i + 1 

  # load configuration file
  conf = configuration.Configuration()
  
  conf.addConfigurable(log.Logger)

  conf.addAllowedKeys(['LOG_INTERVAL', 'LOG_INTERVAL_LIMIT', 'DAEMON_MODE'])
  conf.addAllowedListKeys(['DATASOURCE', 'DATASTORE', 'MEASURE', 'VIRTUAL', 'STOREFILTER'])

  (confstat, errmsg) = conf.loadFile(CONFIGURATION_FILE)
  if confstat == 0:
      print "ERROR: Cannot read configuration file: " + CONFIGURATION_FILE
      sys.exit(1)
  elif confstat == -1:
      print "ERROR: Error reading configuration file: " + CONFIGURATION_FILE
      print errmsg
      sys.exit(1)

  if logconsole:
      conf.setValue('CONSOLE_LOGGING', 'true')
  if verbose:
      conf.setValue('VERBOSE_LOGGING', 'true')

  LOG = log.Logger(conf)
  log.Logging.setLogger(LOG)

  CheckReleasedVersion()  

  sources = []
  for st in conf.getValue('DATASOURCE', []):
      (mtype, mname) = GetModuleTypeAndName(st)
      if len(mtype) <= 0 or len(mname) <= 0:
          LOG.log("ERROR: Error in DATASOURCE definition: " + st)
          sys.exit(1)
          
      sourceClass = GetDataSourceClass(mtype)
      if sourceClass == None:
          LOG.log("ERROR: Invalid datasource type in DATASOURCE definition: " + st)
          sys.exit(1)
      sources.append( sourceClass(mname) )          
                    
  if len(sources) <= 0:
      LOG.log("ERROR: No data sources configured. Nothing to log. Exiting.")
      sys.exit(1)

  stores = []
  for st in conf.getValue('DATASTORE', []):
      (mtype, mname) = GetModuleTypeAndName(st)
      if len(mtype) <= 0 or len(mname) <= 0:
          LOG.log("ERROR: Error in DATASTORE definition: " + st)
          sys.exit(1)
      
      storeClass = GetDataStoreClass(mtype)
      if storeClass == None:
          LOG.log("ERROR: Invalid datastore type in DATASTORE definition: " + st)
          sys.exit(1)
      stores.append( storeClass(mname) )
          
  if len(stores) <= 0:
      LOG.log("WARNING: No log data stores configured. Data is written to application log.")

  if not nodaemon and (daemon or conf.isTrue('DAEMON_MODE')):
      pid = os.fork()
      if (pid != 0):
          LOG.log("Starting taloLogger.py (" + VERSION + ") as daemon.")
          sys.exit(0) 
  else:
      LOG.log("Starting taloLogger.py (" + VERSION + ").")
          
  # try to enable signal handlers, if fails, we are probably using Windows
  try:
      import signal 
      signal.signal(signal.SIGTERM, TERMhandler)
      signal.signal(signal.SIGINT, INThandler)
      signal.signal(signal.SIGHUP, HUPhandler)
  except:
      pass

  LOG_INTERVAL = int(conf.getValue('LOG_INTERVAL', DEFAULT_LOG_INTERVAL))
  LOG_INTERVAL_LIMIT = int(conf.getValue('LOG_INTERVAL_LIMIT', DEFAULT_LOG_INTERVAL_LIMIT))

  measurekeys = []
  measurepositions = []
  measures = []
  for item in conf.getValue('MEASURE', []):
      temps = string.split(item, ':', 1)
      if len(temps) < 2:
          LOG.log("ERROR: Invalid MEASURE configuration: " + item)
          sys.exit(1)          
      key = string.strip(temps[0])
      value = string.strip(temps[1])
      if len(key) <= 0 or len(value) <= 0:
          LOG.log("ERROR: Invalid MEASURE configuration: " + item)
          sys.exit(1)
      if len( string.split(value, '.', 1) ) < 2:
          LOG.log("ERROR: Invalid MEASURE configuration: " + item)
          sys.exit(1)
      if key in measurepositions:
          LOG.log("ERROR: Duplicate MEASURE key: " + key)
          sys.exit(1)                    
      measurepositions.append(key)
      if value not in measurekeys:
          measurekeys.append(value)
      measures.append([key, value])

  if len(measures) <= 0:
      LOG.log("ERROR: No configured measures. Nothing to log. Exiting.")
      sys.exit(1)                    

  virtualpositions = []
  virtuals = []
  for item in conf.getValue('VIRTUAL', []):
      temps = string.split(item, ':', 2)
      if len(temps) < 3:
          LOG.log("ERROR: Invalid VIRTUAL configuration: " + item)
          sys.exit(1)          
      key = string.strip(temps[0])
      try:
          type = int(string.strip(temps[1]))
      except:
          type = 0
      value = string.strip(temps[2])
      if type < 1 or type > 1:
          LOG.log("ERROR: Invalid VIRTUAL configuration type: " + item)
          sys.exit(1)          
      if len(key) <= 0 or len(value) <= 0:
          LOG.log("ERROR: Invalid VIRTUAL configuration: " + item)
          sys.exit(1)
      if key in measurepositions or key in virtualpositions:
          LOG.log("ERROR: Duplicate MEASURE/VIRTUAL key: " + key)
          sys.exit(1)                    
      virtualpositions.append(key)
      virtuals.append([key, type, value])
            
  # handle module configurations
  for confmod in sources + stores:
      (stat, msg) = conf.checkConfigurationKeys(confmod)
      if stat == -1:
          LOG.log("ERROR: Module %s has invalid configuration key: %s" % (confmod.getModuleName(), msg))
          sys.exit(1)          
      elif stat == 0:
          LOG.log("WARNING: Module %s: %s" % (confmod.getModuleName(), msg))
          sys.exit(1)          
            
      (stat, msg) = confmod.handleConfiguration(conf)
      if stat == 0:
          LOG.log("ERROR: Error configuring module " + confmod.getModuleName() + ": " + msg)
      elif stat == -1:
          LOG.log("FATAL ERROR: Error configuring modules. Exiting.: " + msg)
          sys.exit(1)
      elif stat == 1:
          LOG.debug("Module config read " + confmod.getModuleName())

  # initialize module configurations
  for confmod in sources + stores:      
      (stat, msg) = confmod.initConfiguration()
      if stat == 0:
          LOG.log("ERROR: Error initializing module " + confmod.getModuleName() + ": " + msg)
      elif stat == -1:
          LOG.log("FATAL ERROR: Error initializing modules. Exiting.: " + msg)
          sys.exit(1)
      elif stat == 1:
          LOG.log("Initialized module " + confmod.getModuleName())

  # check that STOREFILTER positions are all also measures for active modules
  for store in stores:
      (stat, msg) = store.checkStoreFilters(measurepositions+virtualpositions)
      if stat == 0:
          LOG.log("ERROR: Invalid store filter position for active module: " + msg)
          sys.exit(1)
  
  loggerthread = TaloLoggerThread(measures, virtuals, sources, stores)

  if not loggerthread.start():
      LOG.log("ERROR: Cannot start logger thread.")
      sys.exit(1)

  while not main_terminated and not loggerthread.isterminated():
      t = time.time() % LOG_INTERVAL
      if t < LOG_INTERVAL_LIMIT:
          loggerthread.queue('log', measurekeys, time.time())
          t = time.time() % LOG_INTERVAL
      time.sleep(LOG_INTERVAL - t)

  loggerthread.terminate() 
  loggerthread.wait_sleeping()
  
  for storemod in stores:
      storemod.clearStagedData()
  
  for confmod in sources + stores:
      (stat, msg) = confmod.releaseConfiguration()
      if stat == 0:
          LOG.log("ERROR: Error releasing module: " + msg)

  LOG.log("Terminating.")


def Usage():
    print 'USAGE:  python taloLogger.py --help'
    print '        python taloLogger.py [-d] [--nodaemon] [-v] [-l] [-f configuration_file]'


def Help():
    temps = string.split(HELP_TEXT, '\n')
    for temp in temps:
        if len(temp) > 0 and temp[0] == '#':
            temp = temp[1:]
        if len(temp) > 0 and temp[0] == ' ':
            temp = temp[1:]
        print temp
 

###########################################################################

if __name__ == "__main__":
  main()

